from __future__ import annotations

import importlib
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from .exceptions import ModuleImportError

if TYPE_CHECKING:
    from types import ModuleType

module_mtime_cache: dict[str, float] = {}


def import_module(module_path: Path, import_root: str | None = None, relative_path: str | None = None) -> ModuleType:
    """Imports a python module.

    Args:
        module_path (Path): Absolute path of the module to import.
        import_root (Optional[str]): Absolute string path to the current repository.
        relative_path (Optional[str]): Relative string path between module_path and import_root.
    """
    import_root = import_root or str(module_path.parent)

    file_on_disk = module_path
    if import_root and relative_path:
        file_on_disk = Path(import_root, relative_path, module_path.name)

    # This is a temporary workaround for to account for issues if "import_root" is a Path instead of a str
    # Later we should rework this so that import_root and relative_path are all Path objects. Here we must
    # ensure that anything we add to sys.path is a string and not a Path or PosixPath object.
    import_root = str(import_root)
    if import_root not in sys.path:
        sys.path.append(import_root)

    module_name = module_path.stem

    if relative_path:
        module_name = relative_path.replace("/", ".") + f".{module_name}"

    try:
        if module_name in sys.modules:
            module = sys.modules[module_name]
            current_mtime = file_on_disk.stat().st_mtime

            if module_name in module_mtime_cache:
                last_mtime = module_mtime_cache[module_name]
                if current_mtime == last_mtime:
                    return module

            module_mtime_cache[module_name] = current_mtime
            module = importlib.reload(module)
        else:
            module = importlib.import_module(module_name)
            module_mtime_cache[module_name] = file_on_disk.stat().st_mtime

    except ModuleNotFoundError as exc:
        raise ModuleImportError(message=f"{exc!s} ({module_path})") from exc
    except SyntaxError as exc:
        raise ModuleImportError(message=str(exc)) from exc

    return module



from __future__ import annotations

from copy import copy
from functools import reduce
from typing import TYPE_CHECKING, Any, Iterable, Literal, TypeVar, overload

from infrahub_sdk.utils import deep_merge_dict, is_valid_uuid

from infrahub.core.constants import RelationshipCardinality, RelationshipDirection
from infrahub.core.node import Node
from infrahub.core.node.delete_validator import NodeDeleteValidator
from infrahub.core.query.node import (
    AttributeFromDB,
    AttributeNodePropertyFromDB,
    GroupedPeerNodes,
    NodeAttributesFromDB,
    NodeGetHierarchyQuery,
    NodeGetListQuery,
    NodeListGetAttributeQuery,
    NodeListGetInfoQuery,
    NodeListGetRelationshipsQuery,
    NodeToProcess,
)
from infrahub.core.query.relationship import RelationshipGetPeerQuery
from infrahub.core.registry import registry
from infrahub.core.relationship import Relationship, RelationshipManager
from infrahub.core.schema import (
    GenericSchema,
    MainSchemaTypes,
    NodeSchema,
    ProfileSchema,
    RelationshipSchema,
    TemplateSchema,
)
from infrahub.core.timestamp import Timestamp
from infrahub.exceptions import NodeNotFoundError, ProcessingError, SchemaNotFoundError
from infrahub.graphql.models import OrderModel

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.constants import RelationshipHierarchyDirection
    from infrahub.database import InfrahubDatabase

SchemaProtocol = TypeVar("SchemaProtocol")


def identify_node_class(node: NodeToProcess) -> type[Node]:
    """Identify the proper class to use to create the NodeObject.

    If there is a class in the registry matching the name Kind, use it
    If there is a class in the registry matching one of the node Parent, use it
    Otherwise use Node
    """
    if node.schema.kind in registry.node:
        return registry.node[node.schema.kind]

    if node.schema.inherit_from:
        for parent in node.schema.inherit_from:
            if parent in registry.node:
                return registry.node[parent]

    return Node


def get_schema(
    db: InfrahubDatabase, branch: Branch, node_schema: type[SchemaProtocol] | MainSchemaTypes | str
) -> MainSchemaTypes:
    if isinstance(node_schema, str):
        return db.schema.get(name=node_schema, branch=branch.name)
    if hasattr(node_schema, "_is_runtime_protocol") and node_schema._is_runtime_protocol:
        return db.schema.get(name=node_schema.__name__, branch=branch.name)
    if not isinstance(node_schema, (MainSchemaTypes)):
        raise ValueError(f"Invalid schema provided {node_schema}")

    return node_schema


class ProfileAttributeIndex:
    def __init__(
        self,
        profile_attributes_id_map: dict[str, NodeAttributesFromDB],
        profile_ids_by_node_id: dict[str, list[str]],
    ) -> None:
        self._profile_attributes_id_map = profile_attributes_id_map
        self._profile_ids_by_node_id = profile_ids_by_node_id

    def apply_profiles(self, node_data_dict: dict[str, Any]) -> dict[str, Any]:
        updated_data: dict[str, Any] = {**node_data_dict}
        node_id = node_data_dict.get("id")
        profile_ids = self._profile_ids_by_node_id.get(node_id, [])
        if not profile_ids:
            return updated_data
        profiles = [
            self._profile_attributes_id_map[p_id] for p_id in profile_ids if p_id in self._profile_attributes_id_map
        ]

        def get_profile_priority(nafd: NodeAttributesFromDB) -> tuple[int | float, str]:
            try:
                return (int(nafd.attrs.get("profile_priority").value), nafd.node.get("uuid"))
            except (TypeError, AttributeError):
                return (float("inf"), "")

        profiles.sort(key=get_profile_priority)

        for attr_name, attr_data in updated_data.items():
            if not isinstance(attr_data, AttributeFromDB):
                continue
            if not attr_data.is_default:
                continue
            profile_value, profile_uuid = None, None
            index = 0

            while profile_value is None and index <= (len(profiles) - 1):
                try:
                    profile_value = profiles[index].attrs[attr_name].value
                    if profile_value != "NULL":
                        profile_uuid = profiles[index].node["uuid"]
                        break
                    profile_value = None
                except (IndexError, KeyError, AttributeError):
                    ...
                index += 1

            if profile_value is not None:
                attr_data.value = profile_value
                attr_data.is_from_profile = True
                attr_data.is_default = False
                attr_data.node_properties["source"] = AttributeNodePropertyFromDB(uuid=profile_uuid, labels=[])
        return updated_data


class NodeManager:
    @overload
    @classmethod
    async def query(
        cls,
        db: InfrahubDatabase,
        schema: NodeSchema | GenericSchema | ProfileSchema | TemplateSchema | str,
        filters: dict | None = ...,
        fields: dict | None = ...,
        offset: int | None = ...,
        limit: int | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        partial_match: bool = ...,
        branch_agnostic: bool = ...,
        order: OrderModel | None = ...,
    ) -> list[Any]: ...

    @overload
    @classmethod
    async def query(
        cls,
        db: InfrahubDatabase,
        schema: type[SchemaProtocol],
        filters: dict | None = ...,
        fields: dict | None = ...,
        offset: int | None = ...,
        limit: int | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        partial_match: bool = ...,
        branch_agnostic: bool = ...,
        order: OrderModel | None = ...,
    ) -> list[SchemaProtocol]: ...

    @classmethod
    async def query(
        cls,
        db: InfrahubDatabase,
        schema: type[SchemaProtocol] | MainSchemaTypes | str,
        filters: dict | None = None,
        fields: dict | None = None,
        offset: int | None = None,
        limit: int | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        include_source: bool = False,
        include_owner: bool = False,
        prefetch_relationships: bool = False,
        account=None,
        partial_match: bool = False,
        branch_agnostic: bool = False,
        order: OrderModel | None = None,
    ) -> list[Any]:
        """Query one or multiple nodes of a given type based on filter arguments.

        Args:
            schema (NodeSchema or Str): Infrahub Schema or Name of a schema present in the registry.
            filters (dict, optional): filters provided in a dictionary
            fields (dict, optional): List of fields to include in the response.
            limit (int, optional): Maximum numbers of nodes to return. Defaults to 100.
            at (Timestamp or Str, optional): Timestamp for the query. Defaults to None.
            branch (Branch or Str, optional): Branch to query. Defaults to None.

        Returns:
            list[Node]: List of Node object
        """

        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        node_schema = get_schema(db=db, branch=branch, node_schema=schema)

        if filters and "hfid" in filters:
            node = await cls.get_one_by_hfid(
                db=db,
                hfid=filters["hfid"],
                kind=schema,
                fields=fields,
                at=at,
                branch=branch,
                include_source=include_source,
                include_owner=include_owner,
                prefetch_relationships=prefetch_relationships,
                account=account,
                branch_agnostic=branch_agnostic,
            )
            return [node] if node else []

        # Query the list of nodes matching this Query
        query = await NodeGetListQuery.init(
            db=db,
            schema=node_schema,
            branch=branch,
            offset=offset,
            limit=limit,
            filters=filters,
            at=at,
            partial_match=partial_match,
            branch_agnostic=branch_agnostic,
            order=order,
        )
        await query.execute(db=db)
        node_ids = query.get_node_ids()

        # if display_label or hfid has been requested we need to ensure we are querying the right fields
        if fields and "display_label" in fields:
            schema_branch = db.schema.get_schema_branch(name=branch.name)
            display_label_fields = schema_branch.generate_fields_for_display_label(name=node_schema.kind)
            if display_label_fields:
                fields = deep_merge_dict(dicta=fields, dictb=display_label_fields)

        if fields and "hfid" in fields and node_schema.human_friendly_id:
            hfid_fields = node_schema.generate_fields_for_hfid()
            if hfid_fields:
                fields = deep_merge_dict(dicta=fields, dictb=hfid_fields)

        response = await cls.get_many(
            ids=node_ids,
            fields=fields,
            branch=branch,
            account=account,
            at=at,
            include_source=include_source,
            include_owner=include_owner,
            db=db,
            prefetch_relationships=prefetch_relationships,
            branch_agnostic=branch_agnostic,
        )

        return list(response.values()) if node_ids else []

    @classmethod
    async def count(
        cls,
        db: InfrahubDatabase,
        schema: type[SchemaProtocol] | NodeSchema | GenericSchema | ProfileSchema | TemplateSchema | str,
        filters: dict | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        account=None,  # noqa: ARG003
        partial_match: bool = False,
        branch_agnostic: bool = False,
    ) -> int:
        """Return the total number of nodes using a given filter

        Args:
            schema (NodeSchema): Infrahub Schema or Name of a schema present in the registry.
            filters (dict, optional): filters provided in a dictionary
            at (Timestamp or Str, optional): Timestamp for the query. Defaults to None.
            branch (Branch or Str, optional): Branch to query. Defaults to None.

        Returns:
            int: The number of responses found
        """

        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        node_schema = get_schema(db=db, branch=branch, node_schema=schema)

        query = await NodeGetListQuery.init(
            db=db,
            schema=node_schema,
            branch=branch,
            filters=filters,
            at=at,
            partial_match=partial_match,
            branch_agnostic=branch_agnostic,
            order=OrderModel(disable=True),
        )
        return await query.count(db=db)

    @classmethod
    async def count_peers(
        cls,
        ids: list[str],
        source_kind: str,
        schema: RelationshipSchema,
        filters: dict,
        db: InfrahubDatabase,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        branch_agnostic: bool = False,
    ) -> int:
        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        rel = Relationship(schema=schema, branch=branch, node_id="PLACEHOLDER")

        query = await RelationshipGetPeerQuery.init(
            db=db,
            source_ids=ids,
            source_kind=source_kind,
            schema=schema,
            filters=filters,
            rel=rel,
            at=at,
            branch_agnostic=branch_agnostic,
        )
        return await query.count(db=db)

    @classmethod
    async def query_peers(
        cls,
        db: InfrahubDatabase,
        ids: list[str],
        source_kind: str,
        schema: RelationshipSchema,
        filters: dict,
        fields: dict | None = None,
        offset: int | None = None,
        limit: int | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        branch_agnostic: bool = False,
        fetch_peers: bool = False,
    ) -> list[Relationship]:
        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        rel = Relationship(schema=schema, branch=branch, node_id="PLACEHOLDER")

        query = await RelationshipGetPeerQuery.init(
            db=db,
            source_ids=ids,
            source_kind=source_kind,
            schema=schema,
            filters=filters,
            rel=rel,
            offset=offset,
            limit=limit,
            at=at,
            branch_agnostic=branch_agnostic,
        )
        await query.execute(db=db)

        peers_info = list(query.get_peers())
        if not peers_info:
            return []

        # if display_label has been requested we need to ensure we are querying the right fields
        if fields and "display_label" in fields:
            peer_schema = schema.get_peer_schema(db=db, branch=branch)
            schema_branch = db.schema.get_schema_branch(name=branch.name)
            display_label_fields = schema_branch.generate_fields_for_display_label(name=peer_schema.kind)
            if display_label_fields:
                fields = deep_merge_dict(dicta=fields, dictb=display_label_fields)

        if fields and "hfid" in fields:
            peer_schema = schema.get_peer_schema(db=db, branch=branch)
            hfid_fields = peer_schema.generate_fields_for_hfid()
            if hfid_fields:
                fields = deep_merge_dict(dicta=fields, dictb=hfid_fields)

        if fetch_peers:
            peer_ids = [peer.peer_id for peer in peers_info]
            peer_nodes = await cls.get_many(
                db=db, ids=peer_ids, fields=fields, at=at, branch=branch, branch_agnostic=branch_agnostic
            )

        results = []
        for peer in peers_info:
            result = Relationship(schema=schema, branch=branch, at=at, node_id=peer.source_id).load(
                db=db,
                id=peer.rel_node_id,
                db_id=peer.rel_node_db_id,
                updated_at=peer.updated_at,
                data=peer,
            )
            if fetch_peers:
                result.set_peer(value=peer_nodes[peer.peer_id])
            results.append(result)

        return results

    @classmethod
    async def count_hierarchy(
        cls,
        id: str,
        direction: RelationshipHierarchyDirection,
        node_schema: NodeSchema,
        filters: dict,
        db: InfrahubDatabase,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
    ) -> int:
        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        query = await NodeGetHierarchyQuery.init(
            db=db,
            direction=direction,
            node_id=id,
            node_schema=node_schema,
            filters=filters,
            at=at,
            branch=branch,
        )

        return await query.count(db=db)

    @classmethod
    async def query_hierarchy(
        cls,
        db: InfrahubDatabase,
        id: str,
        direction: RelationshipHierarchyDirection,
        node_schema: NodeSchema,
        filters: dict,
        fields: dict | None = None,
        offset: int | None = None,
        limit: int | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
    ) -> dict[str, Node]:
        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        query = await NodeGetHierarchyQuery.init(
            db=db,
            direction=direction,
            node_id=id,
            node_schema=node_schema,
            filters=filters,
            offset=offset,
            limit=limit,
            at=at,
            branch=branch,
        )
        await query.execute(db=db)

        peers_ids = list(query.get_peer_ids())

        if not peers_ids:
            return {}

        hierarchy_schema = node_schema.get_hierarchy_schema(db=db, branch=branch)

        # if display_label has been requested we need to ensure we are querying the right fields
        if fields and "display_label" in fields:
            schema_branch = db.schema.get_schema_branch(name=branch.name)
            display_label_fields = schema_branch.generate_fields_for_display_label(name=hierarchy_schema.kind)
            if display_label_fields:
                fields = deep_merge_dict(dicta=fields, dictb=display_label_fields)

        return await cls.get_many(
            db=db, ids=peers_ids, fields=fields, at=at, branch=branch, include_owner=True, include_source=True
        )

    @overload
    @classmethod
    async def find_object(
        cls,
        db: InfrahubDatabase,
        kind: type[SchemaProtocol],
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        id: str | None = ...,
        hfid: list[str] | None = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def find_object(
        cls,
        db: InfrahubDatabase,
        kind: str,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        id: str | None = ...,
        hfid: list[str] | None = ...,
    ) -> Any: ...

    @classmethod
    async def find_object(
        cls,
        db: InfrahubDatabase,
        kind: type[SchemaProtocol] | str,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        id: str | None = None,
        hfid: list[str] | None = None,
    ) -> Any:
        if not id and not hfid:
            raise ProcessingError(message="either id or hfid must be provided.")

        if id and is_valid_uuid(id):
            return await cls.get_one(
                db=db,
                kind=kind,
                id=id,
                branch=branch,
                at=at,
                include_owner=True,
                include_source=True,
                raise_on_error=True,
            )

        if hfid:
            return await cls.get_one_by_hfid(
                db=db,
                kind=kind,
                hfid=hfid,
                branch=branch,
                at=at,
                include_owner=True,
                include_source=True,
                raise_on_error=True,
            )

        return await cls.get_one_by_default_filter(
            db=db,
            kind=kind,
            id=id,
            branch=branch,
            at=at,
            include_owner=True,
            include_source=True,
            raise_on_error=True,
        )

    @overload
    @classmethod
    async def get_one_by_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: type[SchemaProtocol],
        raise_on_error: Literal[False] = ...,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol | None: ...

    @overload
    @classmethod
    async def get_one_by_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: type[SchemaProtocol],
        raise_on_error: Literal[True],
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def get_one_by_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: type[SchemaProtocol],
        raise_on_error: bool,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def get_one_by_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: str,
        raise_on_error: bool = ...,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> Any: ...

    @classmethod
    async def get_one_by_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: type[SchemaProtocol] | str,
        raise_on_error: bool = False,
        fields: dict | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        include_source: bool = False,
        include_owner: bool = False,
        prefetch_relationships: bool = False,
        account=None,
        branch_agnostic: bool = False,
    ) -> Any:
        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        node_schema = get_schema(db=db, branch=branch, node_schema=kind)
        kind_str = node_schema.kind

        if not node_schema.default_filter:
            raise NodeNotFoundError(branch_name=branch.name, node_type=kind_str, identifier=id)

        items = await NodeManager.query(
            db=db,
            schema=node_schema,
            fields=fields,
            limit=2,
            filters={node_schema.default_filter: id},
            branch=branch,
            at=at,
            include_owner=include_owner,
            include_source=include_source,
            prefetch_relationships=prefetch_relationships,
            account=account,
            branch_agnostic=branch_agnostic,
            order=OrderModel(disable=True),
        )

        if len(items) > 1:
            raise NodeNotFoundError(
                branch_name=branch.name,
                node_type=kind_str,
                identifier=id,
                message=f"Unable to find node {id!r}, {len(items)} nodes returned, expected 1",
            )

        if items:
            return items[0]
        if not raise_on_error:
            return None

        raise NodeNotFoundError(
            branch_name=branch.name,
            node_type=kind_str,
            identifier=id,
        )

    @overload
    @classmethod
    async def get_one_by_hfid(
        cls,
        db: InfrahubDatabase,
        hfid: list[str],
        kind: type[SchemaProtocol],
        raise_on_error: Literal[False] = ...,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
    ) -> SchemaProtocol | None: ...

    @overload
    @classmethod
    async def get_one_by_hfid(
        cls,
        db: InfrahubDatabase,
        hfid: list[str],
        kind: type[SchemaProtocol],
        raise_on_error: Literal[True],
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def get_one_by_hfid(
        cls,
        db: InfrahubDatabase,
        hfid: list[str],
        kind: type[SchemaProtocol],
        raise_on_error: bool,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def get_one_by_hfid(
        cls,
        db: InfrahubDatabase,
        hfid: list[str],
        kind: str,
        raise_on_error: Literal[True],
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> Any: ...

    @overload
    @classmethod
    async def get_one_by_hfid(
        cls,
        db: InfrahubDatabase,
        hfid: list[str],
        kind: str,
        raise_on_error: bool = ...,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> Any: ...

    @classmethod
    async def get_one_by_hfid(
        cls,
        db: InfrahubDatabase,
        hfid: list[str],
        kind: type[SchemaProtocol] | str,
        raise_on_error: bool = False,
        fields: dict | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        include_source: bool = False,
        include_owner: bool = False,
        prefetch_relationships: bool = False,
        account=None,
        branch_agnostic: bool = False,
    ) -> Any:
        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        node_schema = get_schema(db=db, branch=branch, node_schema=kind)
        kind_str = node_schema.kind

        hfid_str = " :: ".join(hfid)

        if not node_schema.human_friendly_id:
            raise NodeNotFoundError(
                branch_name=branch.name,
                node_type=kind_str,
                identifier=hfid_str,
                message=f"Unable to lookup node by HFID, schema '{node_schema.kind}' does not have a HFID defined.",
            )

        if len(node_schema.human_friendly_id) != len(hfid):
            raise NodeNotFoundError(
                branch_name=branch.name,
                node_type=kind_str,
                identifier=hfid_str,
                message=f"Unable to lookup node by HFID, schema '{node_schema.kind}' HFID does not contain the same number of elements as {hfid}",
            )

        filters = {}
        for key, item in zip(node_schema.human_friendly_id, hfid, strict=False):
            path = node_schema.parse_schema_path(path=key, schema=registry.schema.get_schema_branch(name=branch.name))

            if path.is_type_relationship:
                rel_schema = path.related_schema
                # Keep the relationship attribute path and parse it
                path = rel_schema.parse_schema_path(
                    path=key.split("__", maxsplit=1)[1], schema=registry.schema.get_schema_branch(name=branch.name)
                )

            filters[key] = path.attribute_schema.get_class().deserialize_from_string(item)

        items = await NodeManager.query(
            db=db,
            schema=node_schema,
            fields=fields,
            limit=2,
            filters=filters,
            branch=branch,
            at=at,
            include_owner=include_owner,
            include_source=include_source,
            prefetch_relationships=prefetch_relationships,
            account=account,
            branch_agnostic=branch_agnostic,
            order=OrderModel(disable=True),
        )

        if len(items) < 1:
            if raise_on_error:
                raise NodeNotFoundError(branch_name=branch.name, node_type=kind_str, identifier=hfid_str)
            return None

        if len(items) > 1:
            raise NodeNotFoundError(
                branch_name=branch.name,
                node_type=kind_str,
                identifier=hfid_str,
                message=f"Unable to find node {hfid_str!r}, {len(items)} nodes returned, expected 1",
            )

        return items[0]

    @overload
    @classmethod
    async def get_one_by_id_or_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: type[SchemaProtocol],
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def get_one_by_id_or_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: str,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> Any: ...

    @classmethod
    async def get_one_by_id_or_default_filter(
        cls,
        db: InfrahubDatabase,
        id: str,
        kind: str | type[SchemaProtocol],
        fields: dict | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        include_source: bool = False,
        include_owner: bool = False,
        prefetch_relationships: bool = False,
        account=None,
        branch_agnostic: bool = False,
    ) -> Any:
        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        node = await cls.get_one(
            id=id,
            fields=fields,
            at=at,
            branch=branch,
            include_owner=include_owner,
            include_source=include_source,
            db=db,
            prefetch_relationships=prefetch_relationships,
            account=account,
            branch_agnostic=branch_agnostic,
        )
        if node:
            return node

        node = await cls.get_one_by_default_filter(
            db=db,
            id=id,
            kind=kind,
            fields=fields,
            at=at,
            branch=branch,
            include_source=include_source,
            include_owner=include_owner,
            prefetch_relationships=prefetch_relationships,
            account=account,
            branch_agnostic=branch_agnostic,
        )
        if not node:
            raise NodeNotFoundError(branch_name=branch.name, node_type=kind, identifier=id)
        return node

    @overload
    @classmethod
    async def get_one(
        cls,
        id: str,
        db: InfrahubDatabase,
        kind: type[SchemaProtocol],
        raise_on_error: Literal[False] = ...,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol | None: ...

    @overload
    @classmethod
    async def get_one(
        cls,
        id: str,
        db: InfrahubDatabase,
        kind: type[SchemaProtocol],
        raise_on_error: Literal[True],
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def get_one(
        cls,
        id: str,
        db: InfrahubDatabase,
        kind: type[SchemaProtocol],
        raise_on_error: bool,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> SchemaProtocol: ...

    @overload
    @classmethod
    async def get_one(
        cls,
        id: str,
        db: InfrahubDatabase,
        kind: str,
        raise_on_error: bool = ...,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> Any: ...

    @overload
    @classmethod
    async def get_one(
        cls,
        id: str,
        db: InfrahubDatabase,
        kind: None = ...,
        raise_on_error: bool = ...,
        fields: dict | None = ...,
        at: Timestamp | str | None = ...,
        branch: Branch | str | None = ...,
        include_source: bool = ...,
        include_owner: bool = ...,
        prefetch_relationships: bool = ...,
        account=...,
        branch_agnostic: bool = ...,
    ) -> Any: ...

    @classmethod
    async def get_one(
        cls,
        id: str,
        db: InfrahubDatabase,
        kind: str | type[SchemaProtocol] | None = None,
        raise_on_error: bool = False,
        fields: dict | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        include_source: bool = False,
        include_owner: bool = False,
        prefetch_relationships: bool = False,
        account=None,
        branch_agnostic: bool = False,
    ) -> Any | None:
        """Return one node based on its ID."""
        branch = await registry.get_branch(branch=branch, db=db)

        result = await cls.get_many(
            ids=[id],
            fields=fields,
            at=at,
            branch=branch,
            include_source=include_source,
            include_owner=include_owner,
            account=account,
            prefetch_relationships=prefetch_relationships,
            db=db,
            branch_agnostic=branch_agnostic,
        )

        if not result:
            if raise_on_error:
                raise NodeNotFoundError(branch_name=branch.name, node_type=kind, identifier=id)
            return None

        node = result[id]
        node_schema = node.get_schema()

        kind_validation = None
        if kind:
            node_schema_validation = get_schema(db=db, branch=branch, node_schema=kind)
            kind_validation = node_schema_validation.kind

        # Temporary list of exception to the validation of the kind
        kind_validation_exceptions = [
            ("CoreChangeThread", "CoreObjectThread"),  # issue/3318
        ]

        if kind_validation and (
            node_schema.kind != kind_validation and kind_validation not in node_schema.inherit_from
        ):
            for item in kind_validation_exceptions:
                if item[0] == kind_validation and item[1] == node.get_kind():
                    return node

            raise NodeNotFoundError(
                branch_name=branch.name,
                node_type=kind_validation,
                identifier=id,
                message=f"Node with id {id} exists, but it is a {node.get_kind()}, not {kind_validation}",
            )

        return node

    @classmethod
    async def get_many(
        cls,
        db: InfrahubDatabase,
        ids: list[str],
        fields: dict | None = None,
        at: Timestamp | str | None = None,
        branch: Branch | str | None = None,
        include_source: bool = False,
        include_owner: bool = False,
        prefetch_relationships: bool = False,
        account=None,
        branch_agnostic: bool = False,
    ) -> dict[str, Node]:
        """Return a list of nodes based on their IDs."""

        branch = await registry.get_branch(branch=branch, db=db)
        at = Timestamp(at)

        # Query all nodes
        query = await NodeListGetInfoQuery.init(
            db=db, ids=ids, branch=branch, account=account, at=at, branch_agnostic=branch_agnostic
        )
        await query.execute(db=db)
        nodes_info_by_id: dict[str, NodeToProcess] = {node.node_uuid: node async for node in query.get_nodes(db=db)}
        profile_ids_by_node_id = query.get_profile_ids_by_node_id()
        all_profile_ids = reduce(
            lambda all_ids, these_ids: all_ids | set(these_ids), profile_ids_by_node_id.values(), set()
        )

        if fields and all_profile_ids:
            if "profile_priority" not in fields:
                fields["profile_priority"] = {}
            if "value" not in fields["profile_priority"]:
                fields["profile_priority"]["value"] = None

        # Query list of all Attributes
        query = await NodeListGetAttributeQuery.init(
            db=db,
            ids=list(nodes_info_by_id.keys()) + list(all_profile_ids),
            fields=fields,
            branch=branch,
            include_source=include_source,
            include_owner=include_owner,
            account=account,
            at=at,
            branch_agnostic=branch_agnostic,
        )
        await query.execute(db=db)
        all_node_attributes = query.get_attributes_group_by_node()
        profile_attributes: dict[str, dict[str, AttributeFromDB]] = {}
        node_attributes: dict[str, dict[str, AttributeFromDB]] = {}
        for node_id, attribute_dict in all_node_attributes.items():
            if node_id in all_profile_ids:
                profile_attributes[node_id] = attribute_dict
            else:
                node_attributes[node_id] = attribute_dict
        profile_index = ProfileAttributeIndex(
            profile_attributes_id_map=profile_attributes, profile_ids_by_node_id=profile_ids_by_node_id
        )

        nodes: dict[str, Node] = {}

        for node_id in ids:
            if node_id not in nodes_info_by_id:
                continue

            node = nodes_info_by_id[node_id]
            new_node_data: dict[str, str | AttributeFromDB] = {
                "db_id": node.node_id,
                "id": node_id,
                "updated_at": node.updated_at,
            }

            if not node.schema:
                raise SchemaNotFoundError(
                    branch_name=branch.name,
                    identifier=node_id,
                    message=f"Unable to find the Schema associated with {node_id}, {node.labels}",
                )

            # --------------------------------------------------------
            # Attributes
            # --------------------------------------------------------
            if node_id in node_attributes:
                for attr_name, attr in node_attributes[node_id].attrs.items():
                    new_node_data[attr_name] = attr

            new_node_data_with_profile_overrides = profile_index.apply_profiles(new_node_data)
            node_class = identify_node_class(node=node)
            node_branch = await registry.get_branch(db=db, branch=node.branch)
            item = await node_class.init(schema=node.schema, branch=node_branch, at=at, db=db)
            await item.load(**new_node_data_with_profile_overrides, db=db)

            nodes[node_id] = item

        await cls._enrich_node_dicts_with_relationships(
            db=db,
            branch=branch,
            at=at,
            nodes_by_id=nodes,
            branch_agnostic=branch_agnostic,
            include_owner=include_owner,
            include_source=include_source,
            prefetch_relationships=prefetch_relationships,
            fields=fields,
        )

        return nodes

    @classmethod
    async def _enrich_node_dicts_with_relationships(
        cls,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp,
        nodes_by_id: dict[str, Node],
        branch_agnostic: bool,
        include_owner: bool,
        include_source: bool,
        prefetch_relationships: bool,
        fields: dict[str, Any] | None,
    ) -> None:
        if not prefetch_relationships and not fields:
            return
        cardinality_one_identifiers_by_kind: dict[str, dict[str, RelationshipDirection]] | None = None
        outbound_identifiers: set[str] | None = None
        inbound_identifiers: set[str] | None = None
        bidirectional_identifiers: set[str] | None = None
        if not prefetch_relationships:
            cardinality_one_identifiers_by_kind = _get_cardinality_one_identifiers_by_kind(
                nodes=nodes_by_id.values(), fields=fields or {}
            )
            outbound_identifiers = set()
            inbound_identifiers = set()
            bidirectional_identifiers = set()
            for identifier_direction_map in cardinality_one_identifiers_by_kind.values():
                for identifier, direction in identifier_direction_map.items():
                    if direction is RelationshipDirection.OUTBOUND:
                        outbound_identifiers.add(identifier)
                    elif direction is RelationshipDirection.INBOUND:
                        inbound_identifiers.add(identifier)
                    elif direction is RelationshipDirection.BIDIR:
                        bidirectional_identifiers.add(identifier)

        query = await NodeListGetRelationshipsQuery.init(
            db=db,
            ids=list(nodes_by_id.keys()),
            outbound_identifiers=None if outbound_identifiers is None else list(outbound_identifiers),
            inbound_identifiers=None if inbound_identifiers is None else list(inbound_identifiers),
            bidirectional_identifiers=None if bidirectional_identifiers is None else list(bidirectional_identifiers),
            branch=branch,
            at=at,
            branch_agnostic=branch_agnostic,
        )
        await query.execute(db=db)
        grouped_peer_nodes = query.get_peers_group_by_node()
        peer_ids = grouped_peer_nodes.get_all_peers()
        # there are no peers to enrich the nodes
        if not peer_ids:
            return

        missing_peers: dict[str, Node] = {}
        if prefetch_relationships:
            # only query the peers that are not already part of the main list
            missing_peer_ids = peer_ids - set(nodes_by_id.keys())
            missing_peers = await cls.get_many(
                ids=list(missing_peer_ids),
                branch=branch,
                at=at,
                db=db,
                include_owner=include_owner,
                include_source=include_source,
            )

        for node in nodes_by_id.values():
            await cls._enrich_one_node_with_relationships(
                db=db,
                node=node,
                grouped_peer_nodes=grouped_peer_nodes,
                nodes_by_id=nodes_by_id | missing_peers,
                cardinality_one_identifiers_by_kind=cardinality_one_identifiers_by_kind,
                insert_peer_node=prefetch_relationships,
            )

    @classmethod
    async def _enrich_one_node_with_relationships(
        cls,
        db: InfrahubDatabase,
        node: Node,
        grouped_peer_nodes: GroupedPeerNodes,
        nodes_by_id: dict[str, Node],
        cardinality_one_identifiers_by_kind: dict[str, dict[str, RelationshipDirection]] | None,
        insert_peer_node: bool,
    ) -> None:
        if not grouped_peer_nodes.has_node(node_id=node.get_id()):
            return

        node_schema = node.get_schema()
        for rel_schema in node_schema.relationships:
            peer_ids = grouped_peer_nodes.get_peer_ids(
                node_id=node.get_id(), rel_name=rel_schema.get_identifier(), direction=rel_schema.direction
            )
            if not peer_ids:
                continue

            rel_manager: RelationshipManager = getattr(node, rel_schema.name)
            if insert_peer_node:
                rel_peers: list[Node | str] = []
                for peer_id in peer_ids:
                    peer = nodes_by_id.get(peer_id)
                    if peer:
                        rel_peers.append(peer)
            # if only getting some relationships, make sure we want THIS relationship for THIS node schema
            elif cardinality_one_identifiers_by_kind:
                required_direction = cardinality_one_identifiers_by_kind.get(node_schema.kind, {}).get(
                    rel_schema.get_identifier()
                )
                if required_direction is not rel_schema.direction:
                    continue
                rel_peers = list(peer_ids)
            else:
                continue
            if rel_schema.cardinality is RelationshipCardinality.ONE and len(rel_peers) > 1:
                raise ValueError("At most, one relationship expected")

            rel_manager.has_fetched_relationships = True
            await rel_manager.update(db=db, data=rel_peers)

    @classmethod
    async def delete(
        cls,
        db: InfrahubDatabase,
        nodes: list[Node],
        branch: Branch | str | None = None,
        at: Timestamp | str | None = None,
        cascade_delete: bool = True,
    ) -> list[Node]:
        """Returns list of deleted nodes because of cascading deletes"""
        branch = await registry.get_branch(branch=branch, db=db)
        nodes_to_delete = copy(nodes)
        if cascade_delete:
            node_delete_validator = NodeDeleteValidator(db=db, branch=branch)
            ids_to_delete = await node_delete_validator.get_ids_to_delete(nodes=nodes, at=at)
            node_ids = {node.get_id() for node in nodes}
            missing_ids_to_delete = ids_to_delete - node_ids
            if missing_ids_to_delete:
                node_map = await cls.get_many(db=db, ids=list(missing_ids_to_delete), branch=branch, at=at)
                nodes_to_delete += list(node_map.values())

        for node in nodes_to_delete:
            await node.delete(db=db, at=at)

        return nodes_to_delete


def _get_cardinality_one_identifiers_by_kind(
    nodes: Iterable[Node],
    fields: dict[str, Any],
) -> dict[str, dict[str, RelationshipDirection]]:
    # {kind: {relationship_identifier, ...}}
    cardinality_one_fields_by_kind = {}
    field_names_set = set(fields.keys())
    for node in nodes:
        node_schema = node.get_schema()
        if not node_schema:
            continue
        # already handled this schema
        if node_schema.kind in cardinality_one_fields_by_kind:
            continue
        cardinality_one_rel_identifiers_in_fields = {
            rel_schema.identifier: rel_schema.direction
            for rel_schema in node_schema.relationships
            if rel_schema.cardinality is RelationshipCardinality.ONE and rel_schema.name in field_names_set
        }
        cardinality_one_fields_by_kind[node_schema.kind] = cardinality_one_rel_identifiers_in_fields
    return cardinality_one_fields_by_kind


registry.manager = NodeManager



from __future__ import annotations

from typing import Any, Sequence

from ..query.node_duplicate import NodeDuplicateQuery, SchemaNodeInfo
from ..shared import MigrationQuery, SchemaMigration


class NodeKindUpdateMigrationQuery01(MigrationQuery, NodeDuplicateQuery):
    name = "migration_node_kind_update_01"

    def __init__(
        self,
        migration: SchemaMigration,
        **kwargs: Any,
    ):
        new_node = SchemaNodeInfo(
            name=migration.new_schema.name,
            namespace=migration.new_schema.namespace,
            branch_support=migration.new_schema.branch.value,
            labels=migration.new_schema.get_labels(),
            kind=migration.new_schema.kind,
        )
        previous_node = SchemaNodeInfo(
            name=migration.previous_schema.name,
            namespace=migration.previous_schema.namespace,
            branch_support=migration.previous_schema.branch.value,
            labels=migration.previous_schema.get_labels(),
            kind=migration.previous_schema.kind,
        )
        super().__init__(migration=migration, new_node=new_node, previous_node=previous_node, **kwargs)

    def get_nbr_migrations_executed(self) -> int:
        return self.stats.get_counter(name="nodes_created")


class NodeKindUpdateMigration(SchemaMigration):
    name: str = "node.kind.update"
    queries: Sequence[type[MigrationQuery]] = [NodeKindUpdateMigrationQuery01]  # type: ignore[assignment]



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING, Sequence

from infrahub.core.branch.models import Branch
from infrahub.core.initialization import initialization
from infrahub.core.ipam.reconciler import IpamReconciler
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.timestamp import Timestamp
from infrahub.lock import initialize_lock
from infrahub.log import get_logger

from ..shared import InternalSchemaMigration, SchemaMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class Migration026(InternalSchemaMigration):
    name: str = "026_prefix_0000_fix"
    minimum_version: int = 25
    migrations: Sequence[SchemaMigration] = []

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        # load schemas from database into registry
        initialize_lock()
        await initialization(db=db)

        at = Timestamp()
        for branch in await Branch.get_list(db=db):
            prefix_0000s = await NodeManager.query(
                db=db, schema="BuiltinIPPrefix", branch=branch, filters={"prefix__values": ["0.0.0.0/0", "::/0"]}
            )
            if not prefix_0000s:
                continue
            ipam_reconciler = IpamReconciler(db=db, branch=branch)
            for prefix in prefix_0000s:
                ip_namespace = await prefix.ip_namespace.get_peer(db=db)
                ip_network = ipaddress.ip_network(prefix.prefix.value)
                await ipam_reconciler.reconcile(
                    ip_value=ip_network,
                    namespace=ip_namespace,
                    node_uuid=prefix.get_id(),
                    at=at,
                )

        return MigrationResult()



from __future__ import annotations

from typing import Any, Sequence

from ..query import AttributeMigrationQuery
from ..query.attribute_rename import AttributeInfo, AttributeRenameQuery
from ..shared import AttributeSchemaMigration


class AttributeNameUpdateMigrationQuery01(AttributeMigrationQuery, AttributeRenameQuery):
    name = "migration_attribute_name_update_01"

    def __init__(
        self,
        migration: AttributeSchemaMigration,
        **kwargs: Any,
    ):
        new_attr = AttributeInfo(
            name=migration.new_attribute_schema.name,
            node_kind=migration.new_schema.kind,
            branch_support=migration.new_attribute_schema.get_branch().value,
        )

        attr_id = migration.new_attribute_schema.id
        if not attr_id:
            raise ValueError("The Id is not defined on new_attribute_schema")
        prev_attr_schema = migration.previous_schema.get_attribute_by_id(id=attr_id)

        previous_attr = AttributeInfo(
            name=prev_attr_schema.name,
            node_kind=migration.previous_schema.kind,
            branch_support=prev_attr_schema.get_branch().value,
        )

        super().__init__(migration=migration, new_attr=new_attr, previous_attr=previous_attr, **kwargs)

    def get_nbr_migrations_executed(self) -> int:
        return self.stats.get_counter(name="nodes_created")


class AttributeNameUpdateMigration(AttributeSchemaMigration):
    name: str = "attribute.name.update"
    queries: Sequence[type[AttributeMigrationQuery]] = [AttributeNameUpdateMigrationQuery01]  # type: ignore[assignment]



from infrahub.message_bus import messages
from infrahub.services import InfrahubServices
from infrahub.tasks.registry import refresh_branches
from infrahub.worker import WORKER_IDENTITY


async def branches(message: messages.RefreshRegistryBranches, service: InfrahubServices) -> None:
    if message.meta and message.meta.initiator_id == WORKER_IDENTITY:
        service.log.info("Ignoring refresh registry refresh request originating from self", worker=WORKER_IDENTITY)
        return

    async with service.database.start_session(read_only=True) as db:
        await refresh_branches(db=db)

    await service.component.refresh_schema_hash()


async def rebased_branch(message: messages.RefreshRegistryRebasedBranch, service: InfrahubServices) -> None:
    if message.meta and message.meta.initiator_id == WORKER_IDENTITY:
        service.log.info(
            "Ignoring refresh registry refreshed branch for request originating from self", worker=WORKER_IDENTITY
        )
        return

    async with service.database.start_session(read_only=True) as db:
        await refresh_branches(db=db)

    await service.component.refresh_schema_hash()



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from infrahub.core import registry
from infrahub.core.migrations.shared import MigrationResult
from infrahub.core.schema.definitions.core import core_profile_schema_definition
from infrahub.core.schema.manager import SchemaManager
from infrahub.log import get_logger

from ..shared import InternalSchemaMigration, SchemaMigration

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class Migration017(InternalSchemaMigration):
    name: str = "017_add_graph_migration"
    minimum_version: int = 16
    migrations: Sequence[SchemaMigration] = []

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()

        return result

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        """
        Load CoreProfile schema node in db.
        """
        default_branch = registry.get_branch_from_registry()
        manager = SchemaManager()
        manager.set_schema_branch(name=default_branch.name, schema=self.get_internal_schema())

        db.add_schema(manager.get_schema_branch(default_branch.name))
        await manager.load_node_to_db(node=core_profile_schema_definition, db=db, branch=default_branch)

        return MigrationResult()



from __future__ import annotations

import importlib
import inspect
from typing import TYPE_CHECKING, Any

from prefect import Flow

from infrahub.context import InfrahubContext

if TYPE_CHECKING:
    from infrahub.services import InfrahubServices


def inject_service_parameter(func: Flow, parameters: dict[str, Any], service: InfrahubServices) -> None:
    """
    `service` object instantiates connections to various services (db, cache...) at worker startup,
    so it is not meant to be sent by the server payload. We inject it here to avoid relying on a global variable.
    This mutates input `parameters`.
    """

    # avoid circular imports
    from infrahub.services import InfrahubServices  # pylint: disable=C0415

    if service_parameter_name := get_parameter_name(func=func, types=[InfrahubServices.__name__, InfrahubServices]):
        if any(isinstance(param_value, InfrahubServices) for param_value in parameters):
            raise ValueError(f"{func.name} parameters contains an InfrahubServices object while it should be injected")
        parameters[service_parameter_name] = service
        return


def inject_context_parameter(func: Flow, parameters: dict[str, Any], context: InfrahubContext | None = None) -> None:
    service_parameter_name = get_parameter_name(func=func, types=[InfrahubContext.__name__, InfrahubContext])
    if service_parameter_name and context:
        parameters[service_parameter_name] = context
        return

    if service_parameter_name and not context:
        raise ValueError(
            f"{func.name} has a {service_parameter_name} parameter of type InfrahubContext, while context is not provided"
        )


def load_flow_function(module_path: str, flow_name: str) -> Flow:
    module = importlib.import_module(module_path)
    flow_func = getattr(module, flow_name)
    if not isinstance(flow_func, Flow):
        raise ValueError(
            f"Function loaded at {module_path=} with {flow_name=} has type {type(flow_func)}, expected {Flow}"
        )
    return flow_func


def get_parameter_name(func: Flow, types: list[Any]) -> str | None:
    sig = inspect.signature(func)
    for sig_param in sig.parameters.values():
        if sig_param.annotation in types:
            return sig_param.name
    return None


def has_parameter(func: Flow, types: list[Any]) -> bool:
    return get_parameter_name(func=func, types=types) is not None



from __future__ import annotations

import ipaddress
from typing import TYPE_CHECKING

from graphene import Field, Int, ObjectType, String
from netaddr import IPSet

from infrahub.core.constants import InfrahubKind
from infrahub.core.manager import NodeManager
from infrahub.core.query.ipam import get_ip_addresses, get_subnets
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.pools.address import get_available
from infrahub.pools.prefix import get_next_available_prefix

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class IPAddressGetNextAvailable(ObjectType):
    address = String(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        prefix_id: str,
        prefix_length: int | None = None,
    ) -> dict[str, str]:
        graphql_context: GraphqlContext = info.context

        prefix = await NodeManager.get_one(id=prefix_id, db=graphql_context.db, branch=graphql_context.branch)

        if not prefix:
            raise NodeNotFoundError(
                branch_name=graphql_context.branch.name, node_type=InfrahubKind.IPPREFIX, identifier=prefix_id
            )

        ip_prefix = ipaddress.ip_network(prefix.prefix.value)  # type: ignore[attr-defined]
        prefix_length = prefix_length or ip_prefix.prefixlen

        if not ip_prefix.prefixlen <= prefix_length <= ip_prefix.max_prefixlen:
            raise ValidationError(input_value="Invalid prefix length for current selected prefix")

        namespace = await prefix.ip_namespace.get_peer(db=graphql_context.db)  # type: ignore[attr-defined]
        addresses = await get_ip_addresses(
            db=graphql_context.db,
            ip_prefix=ip_prefix,
            namespace=namespace,
            branch=graphql_context.branch,
        )

        available = get_available(
            network=ip_prefix,
            addresses=[ip.address for ip in addresses],
            is_pool=prefix.is_pool.value,  # type: ignore[attr-defined]
        )

        if not available:
            raise IndexError("No addresses available in prefix")

        next_address = available.iter_cidrs()[0]

        return {"address": f"{next_address.ip}/{prefix_length}"}


class IPPrefixGetNextAvailable(ObjectType):
    prefix = String(required=True)

    @staticmethod
    async def resolve(
        root: dict,  # noqa: ARG004
        info: GraphQLResolveInfo,
        prefix_id: str,
        prefix_length: int,
    ) -> dict[str, str]:
        graphql_context: GraphqlContext = info.context

        prefix = await NodeManager.get_one(id=prefix_id, db=graphql_context.db, branch=graphql_context.branch)

        if not prefix:
            raise NodeNotFoundError(
                branch_name=graphql_context.branch.name, node_type=InfrahubKind.IPPREFIX, identifier=prefix_id
            )

        namespace = await prefix.ip_namespace.get_peer(db=graphql_context.db)  # type: ignore[attr-defined]
        subnets = await get_subnets(
            db=graphql_context.db,
            ip_prefix=ipaddress.ip_network(prefix.prefix.value),  # type: ignore[attr-defined]
            namespace=namespace,
            branch=graphql_context.branch,
        )

        pool = IPSet([prefix.prefix.value])
        for subnet in subnets:
            pool.remove(addr=str(subnet.prefix))

        prefix_ver = ipaddress.ip_network(prefix.prefix.value).version
        next_available = get_next_available_prefix(pool=pool, prefix_length=prefix_length, prefix_ver=prefix_ver)

        return {"prefix": str(next_available)}


InfrahubIPAddressGetNextAvailable = Field(
    IPAddressGetNextAvailable,
    prefix_id=String(required=True),
    prefix_length=Int(required=False),
    resolver=IPAddressGetNextAvailable.resolve,
    required=True,
)


InfrahubIPPrefixGetNextAvailable = Field(
    IPPrefixGetNextAvailable,
    prefix_id=String(required=True),
    prefix_length=Int(required=False),
    resolver=IPPrefixGetNextAvailable.resolve,
    required=True,
)



from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from rich import print as rprint
from rich.console import Console
from rich.table import Table

from infrahub.core import registry
from infrahub.core.branch.models import Branch
from infrahub.core.constants import InfrahubKind
from infrahub.core.migrations.query.node_duplicate import NodeDuplicateQuery, SchemaNodeInfo
from infrahub.core.query import Query, QueryType
from infrahub.core.schema import SchemaRoot, internal_schema
from infrahub.core.schema.manager import SchemaManager
from infrahub.log import get_logger

from ..constants import FAILED_BADGE, SUCCESS_BADGE

if TYPE_CHECKING:
    from infrahub.core.schema.node_schema import NodeSchema
    from infrahub.database import InfrahubDatabase

log = get_logger()


class GetSchemaWithUpdatedInheritance(Query):
    """
    Get the name, namespace, and branch of any SchemaNodes with _updated_ inheritance
    This query will only return schemas that have had `inherit_from` updated after they were created
    """

    name = "get_schema_with_updated_inheritance"
    type = QueryType.READ
    insert_return = False

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        query = """
// find inherit_from attributes that have been updated
MATCH p = (schema_node:SchemaNode)-[has_attr_e:HAS_ATTRIBUTE {status: "active"}]->(a:Attribute {name: "inherit_from"})
WHERE has_attr_e.to IS NULL
CALL (a) {
  // only get branches on which the value was updated, we can ignore the initial create
  MATCH (a)-[e:HAS_VALUE]->(:AttributeValue)
  ORDER BY e.from ASC
  // tail leaves out the earliest one, which is the initial create
  RETURN tail(collect(e.branch)) AS branches
}
WITH schema_node, a, branches
WHERE size(branches) > 0
UNWIND branches AS branch
WITH DISTINCT schema_node, a, branch

//get branch details
CALL (branch) {
  MATCH (b:Branch {name: branch})
  RETURN b.branched_from AS branched_from, b.hierarchy_level AS branch_level
}

// get the namespace for the schema
CALL (schema_node, a, branch, branched_from, branch_level) {
  MATCH (schema_node)-[e1:HAS_ATTRIBUTE]-(:Attribute {name: "namespace"})-[e2:HAS_VALUE]->(av)
  WHERE (
    e1.branch = branch OR
    (e1.branch_level < branch_level AND e1.from <= branched_from)
  ) AND e1.to IS NULL
  AND e1.status = "active"
  AND (
    e2.branch = branch OR
    (e2.branch_level < branch_level AND e2.from <= branched_from)
  ) AND e2.to IS NULL
  AND e2.status = "active"
  ORDER BY e2.branch_level DESC, e1.branch_level DESC, e2.from DESC, e1.from DESC
  RETURN av.value AS namespace
  LIMIT 1
}

// get the name for the schema
CALL (schema_node, a, branch, branched_from, branch_level) {
  MATCH (schema_node)-[e1:HAS_ATTRIBUTE]-(:Attribute {name: "name"})-[e2:HAS_VALUE]->(av)
  WHERE (
    e1.branch = branch OR
    (e1.branch_level < branch_level AND e1.from <= branched_from)
  ) AND e1.to IS NULL
  AND e1.status = "active"
  AND (
    e2.branch = branch OR
    (e2.branch_level < branch_level AND e2.from <= branched_from)
  ) AND e2.to IS NULL
  AND e2.status = "active"
  ORDER BY e2.branch_level DESC, e1.branch_level DESC, e2.from DESC, e1.from DESC
  RETURN av.value AS name
  LIMIT 1
}
RETURN name, namespace, branch
"""
        self.return_labels = ["name", "namespace", "branch"]
        self.add_to_query(query)

    def get_updated_inheritance_kinds_by_branch(self) -> dict[str, list[str]]:
        kinds_by_branch: dict[str, list[str]] = defaultdict(list)
        for result in self.results:
            name = result.get_as_type(label="name", return_type=str)
            namespace = result.get_as_type(label="namespace", return_type=str)
            branch = result.get_as_type(label="branch", return_type=str)
            kinds_by_branch[branch].append(f"{namespace}{name}")
        return kinds_by_branch


@dataclass
class KindLabelCount:
    kind: str
    labels: frozenset[str]
    num_nodes: int


@dataclass
class KindLabelCountCorrected(KindLabelCount):
    node_schema: NodeSchema


class GetAllKindsAndLabels(Query):
    """
    Get the kind, labels, and number of nodes for the given kinds and branch
    """

    name = "get_all_kinds_and_labels"
    type = QueryType.READ
    insert_return = False

    def __init__(self, kinds: list[str] | None = None, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.kinds = kinds

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["branch_name"] = self.branch.name
        self.params["branched_from"] = self.branch.get_branched_from()
        self.params["branch_level"] = self.branch.hierarchy_level
        kinds_str = "Node"
        if self.kinds:
            kinds_str = "|".join(self.kinds)
        query = """
MATCH (n:%(kinds_str)s)-[r:IS_PART_OF]->(:Root)
WHERE (
    r.branch = $branch_name OR
    (r.branch_level < $branch_level AND r.from <= $branched_from)
)
AND r.to IS NULL
AND r.status = "active"
RETURN DISTINCT n.kind AS kind, labels(n) AS labels, count(*) AS num_nodes
ORDER BY kind ASC
        """ % {"kinds_str": kinds_str}
        self.return_labels = ["kind", "labels", "num_nodes"]
        self.add_to_query(query)

    def get_kind_label_counts(self) -> list[KindLabelCount]:
        kind_label_counts: list[KindLabelCount] = []
        for result in self.results:
            kind = result.get_as_type(label="kind", return_type=str)
            num_nodes = result.get_as_type(label="num_nodes", return_type=int)
            labels: list[str] = result.get_as_type(label="labels", return_type=list)
            # we can ignore the Node label and the label that matches the kind
            cleaned_labels = frozenset(str(lbl) for lbl in labels if lbl not in ["Node", "CoreNode", kind])
            kind_label_counts.append(KindLabelCount(kind=kind, labels=cleaned_labels, num_nodes=num_nodes))
        return kind_label_counts


def display_kind_label_counts(kind_label_counts_by_branch: dict[str, list[KindLabelCountCorrected]]) -> None:
    console = Console()

    table = Table(title="Incorrect Inheritance Nodes")

    table.add_column("Branch")
    table.add_column("Kind")
    table.add_column("Incorrect Labels")
    table.add_column("Num Nodes")

    for branch_name, kind_label_counts in kind_label_counts_by_branch.items():
        for kind_label_count in kind_label_counts:
            table.add_row(
                branch_name, kind_label_count.kind, str(list(kind_label_count.labels)), str(kind_label_count.num_nodes)
            )

    console.print(table)


async def check_inheritance(db: InfrahubDatabase, fix: bool = False) -> bool:
    """
    Run migrations to update the inheritance of any nodes with incorrect inheritance from a failed migration
    1. Identifies node schemas that have had their inheritance updated after they were created
        a. includes the kind and branch of the inheritance update
    2. Checks nodes of the given kinds on the given branch to verify their inheritance is correct
    3. Displays counts of any kinds with incorrect inheritance on the given branch
    4. If fix is True, runs migrations to update the inheritance of any nodes with incorrect inheritance
        on the correct branch
    """

    updated_inheritance_query = await GetSchemaWithUpdatedInheritance.init(db=db)
    await updated_inheritance_query.execute(db=db)
    updated_inheritance_kinds_by_branch = updated_inheritance_query.get_updated_inheritance_kinds_by_branch()

    if not updated_inheritance_kinds_by_branch:
        rprint(f"{SUCCESS_BADGE} No schemas have had their inheritance updated")
        return True

    schema_manager = SchemaManager()
    registry.schema = schema_manager
    schema = SchemaRoot(**internal_schema)
    schema_manager.register_schema(schema=schema)
    branches_by_name = {b.name: b for b in await Branch.get_list(db=db)}

    kind_label_counts_by_branch: dict[str, list[KindLabelCountCorrected]] = defaultdict(list)
    for branch_name, kinds in updated_inheritance_kinds_by_branch.items():
        rprint(f"Checking branch: {branch_name}", end="...")
        branch = branches_by_name[branch_name]
        schema_branch = await schema_manager.load_schema_from_db(db=db, branch=branch)
        kind_label_query = await GetAllKindsAndLabels.init(db=db, branch=branch, kinds=kinds)
        await kind_label_query.execute(db=db)
        kind_label_counts = kind_label_query.get_kind_label_counts()

        for kind_label_count in kind_label_counts:
            node_schema = schema_branch.get_node(name=kind_label_count.kind, duplicate=False)
            correct_labels = frozenset(node_schema.inherit_from)
            if kind_label_count.labels == correct_labels:
                continue

            kind_label_counts_by_branch[branch_name].append(
                KindLabelCountCorrected(
                    kind=kind_label_count.kind,
                    labels=kind_label_count.labels,
                    num_nodes=kind_label_count.num_nodes,
                    node_schema=node_schema,
                )
            )
        rprint("done")

    if not kind_label_counts_by_branch:
        rprint(f"{SUCCESS_BADGE} All nodes have the correct inheritance")
        return True

    display_kind_label_counts(kind_label_counts_by_branch)

    if not fix:
        rprint(f"{FAILED_BADGE} Use the --fix flag to fix the inheritance of any invalid nodes")
        return False

    for branch_name, kind_label_counts_corrected in kind_label_counts_by_branch.items():
        for kind_label_count in kind_label_counts_corrected:
            rprint(f"Fixing kind {kind_label_count.kind} on branch {branch_name}", end="...")
            node_schema = kind_label_count.node_schema
            migration_query = await NodeDuplicateQuery.init(
                db=db,
                branch=branches_by_name[branch_name],
                previous_node=SchemaNodeInfo(
                    name=node_schema.name,
                    namespace=node_schema.namespace,
                    branch_support=node_schema.branch.value,
                    labels=list(kind_label_count.labels) + [kind_label_count.kind, InfrahubKind.NODE],
                    kind=kind_label_count.kind,
                ),
                new_node=SchemaNodeInfo(
                    name=node_schema.name,
                    namespace=node_schema.namespace,
                    branch_support=node_schema.branch.value,
                    labels=list(node_schema.inherit_from) + [kind_label_count.kind, InfrahubKind.NODE],
                    kind=kind_label_count.kind,
                ),
            )
            await migration_query.execute(db=db)
            rprint("done")

    rprint(f"{SUCCESS_BADGE} All nodes have the correct inheritance")

    if registry.default_branch in kind_label_counts_by_branch:
        kinds = [kind_label_count.kind for kind_label_count in kind_label_counts_by_branch[registry.default_branch]]
        rprint(
            "[bold cyan]Note that migrations were run on the default branch for the following schema kinds: "
            f"{', '.join(kinds)}. You should rebase any branches that include/will include changes using "
            "the migrated schemas[/bold cyan]"
        )

    return True



from __future__ import annotations

from typing import TYPE_CHECKING, Any, cast

from graphene import InputObjectType, Mutation
from typing_extensions import Self

from infrahub.core.protocols import CoreNodeTriggerAttributeMatch, CoreNodeTriggerRelationshipMatch, CoreNodeTriggerRule
from infrahub.exceptions import SchemaNotFoundError, ValidationError
from infrahub.log import get_logger

from .main import InfrahubMutationMixin, InfrahubMutationOptions

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.schema import NodeSchema
    from infrahub.database import InfrahubDatabase

    from ..initialization import GraphqlContext

log = get_logger()


class InfrahubTriggerRuleMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        _validate_node_kind(data=data, db=db)
        trigger_rule_definition, result = await super().mutate_create(info=info, data=data, branch=branch, database=db)

        return trigger_rule_definition, result

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,  # noqa: ARG003
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        _validate_node_kind(data=data, db=db)
        trigger_rule_definition, result = await super().mutate_update(info=info, data=data, branch=branch, database=db)

        return trigger_rule_definition, result


class InfrahubTriggerRuleMatchMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema,
        _meta: Any | None = None,
        **options: dict[str, Any],
    ) -> None:
        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context

        async with graphql_context.db.start_transaction() as dbt:
            trigger_match, result = await super().mutate_create(info=info, data=data, branch=branch, database=dbt)
            trigger_match_model = cast(CoreNodeTriggerAttributeMatch | CoreNodeTriggerRelationshipMatch, trigger_match)
            node_trigger_rule = await trigger_match_model.trigger.get_peer(db=dbt, raise_on_error=True)
            node_trigger_rule_model = cast(CoreNodeTriggerRule, node_trigger_rule)
            node_schema = dbt.schema.get_node_schema(name=node_trigger_rule_model.node_kind.value, duplicate=False)
            _validate_node_kind_field(data=data, node_schema=node_schema)

        return trigger_match, result

    @classmethod
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,  # noqa: ARG003
        node: Node | None = None,  # noqa: ARG003
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        async with graphql_context.db.start_transaction() as dbt:
            trigger_match, result = await super().mutate_update(info=info, data=data, branch=branch, database=dbt)
            trigger_match_model = cast(CoreNodeTriggerAttributeMatch | CoreNodeTriggerRelationshipMatch, trigger_match)
            node_trigger_rule = await trigger_match_model.trigger.get_peer(db=dbt, raise_on_error=True)
            node_trigger_rule_model = cast(CoreNodeTriggerRule, node_trigger_rule)
            node_schema = dbt.schema.get_node_schema(name=node_trigger_rule_model.node_kind.value, duplicate=False)
            _validate_node_kind_field(data=data, node_schema=node_schema)

        return trigger_match, result


def _validate_node_kind(data: InputObjectType, db: InfrahubDatabase) -> None:
    input_data = cast(dict[str, dict[str, Any]], data)
    if node_kind := input_data.get("node_kind"):
        value = node_kind.get("value")
        if isinstance(value, str):
            try:
                db.schema.get_node_schema(name=value, duplicate=False)
            except SchemaNotFoundError as exc:
                raise ValidationError(
                    input_value={"node_kind": "The requested node_kind schema was not found"}
                ) from exc
            except ValueError as exc:
                raise ValidationError(input_value={"node_kind": "The requested node_kind is not a valid node"}) from exc


def _validate_node_kind_field(data: InputObjectType, node_schema: NodeSchema) -> None:
    input_data = cast(dict[str, dict[str, Any]], data)
    if attribute_name := input_data.get("attribute_name"):
        value = attribute_name.get("value")
        if isinstance(value, str):
            if value not in node_schema.attribute_names:
                raise ValidationError(
                    input_value={
                        "attribute_name": f"The attribute {value} doesn't exist on related node trigger using {node_schema.kind}"
                    }
                )
    if relationship_name := input_data.get("relationship_name"):
        value = relationship_name.get("value")
        if isinstance(value, str):
            if value not in node_schema.relationship_names:
                raise ValidationError(
                    input_value={
                        "relationship_name": f"The relationship {value} doesn't exist on related node trigger using {node_schema.kind}"
                    }
                )



import os
import shutil
import time
import uuid
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path

from testcontainers.compose import DockerCompose
from testcontainers.core.exceptions import ContainerIsNotRunning
from typing_extensions import Self

from infrahub_testcontainers import __version__ as infrahub_version


@dataclass
class ContainerService:
    container: str
    port: int


INFRAHUB_SERVICES: dict[str, ContainerService] = {
    "server": ContainerService(container="infrahub-server-lb", port=8000),
    "task-manager": ContainerService(container="task-manager", port=4200),
    "database": ContainerService(container="database", port=7687),
    "scraper": ContainerService(container="scraper", port=8428),
    "cadvisor": ContainerService(container="cadvisor", port=8080),
}

PROJECT_ENV_VARIABLES: dict[str, str] = {
    "MESSAGE_QUEUE_DOCKER_IMAGE": "rabbitmq:3.13.7-management",
    "CACHE_DOCKER_IMAGE": "redis:7.2.4",
    "INFRAHUB_TESTING_DOCKER_IMAGE": "registry.opsmill.io/opsmill/infrahub",
    "INFRAHUB_TESTING_DOCKER_ENTRYPOINT": f"gunicorn --config backend/infrahub/serve/gunicorn_config.py -w {os.environ.get('INFRAHUB_TESTING_WEB_CONCURRENCY', 4)} --logger-class infrahub.serve.log.GunicornLogger infrahub.server:app",  # noqa: E501
    "INFRAHUB_TESTING_IMAGE_VERSION": infrahub_version,
    "INFRAHUB_TESTING_PRODUCTION": "false",
    "INFRAHUB_TESTING_DB_ADDRESS": "database",
    "INFRAHUB_TESTING_LOG_LEVEL": "INFO",
    "INFRAHUB_TESTING_GIT_REPOSITORIES_DIRECTORY": "/opt/infrahub/git",
    "INFRAHUB_TESTING_API_TOKEN": "44af444d-3b26-410d-9546-b758657e026c",
    "INFRAHUB_TESTING_INITIAL_ADMIN_TOKEN": "06438eb2-8019-4776-878c-0941b1f1d1ec",
    "INFRAHUB_TESTING_INITIAL_AGENT_TOKEN": "44af444d-3b26-410d-9546-b758657e026c",
    "INFRAHUB_TESTING_SECURITY_SECRET_KEY": "327f747f-efac-42be-9e73-999f08f86b92",
    "INFRAHUB_TESTING_ADDRESS": "http://infrahub-server-lb:8000",
    "INFRAHUB_TESTING_INTERNAL_ADDRESS": "http://infrahub-server-lb:8000",
    "INFRAHUB_TESTING_BROKER_ADDRESS": "message-queue",
    "INFRAHUB_TESTING_CACHE_ADDRESS": "cache",
    "INFRAHUB_TESTING_WORKFLOW_ADDRESS": "task-manager",
    "INFRAHUB_TESTING_WORKFLOW_DEFAULT_WORKER_TYPE": "infrahubasync",
    "INFRAHUB_TESTING_TIMEOUT": "60",
    "INFRAHUB_TESTING_PREFECT_API": "http://task-manager:4200/api",
    "INFRAHUB_TESTING_LOCAL_REMOTE_GIT_DIRECTORY": "repos",
    "INFRAHUB_TESTING_INTERNAL_REMOTE_GIT_DIRECTORY": "/remote",
    "INFRAHUB_TESTING_WEB_CONCURRENCY": "4",
    "INFRAHUB_TESTING_LOCAL_DB_BACKUP_DIRECTORY": "backups",
    "INFRAHUB_TESTING_INTERNAL_DB_BACKUP_DIRECTORY": "/backups",
    "INFRAHUB_TESTING_API_SERVER_COUNT": "2",
    "INFRAHUB_TESTING_TASK_WORKER_COUNT": "2",
    "INFRAHUB_TESTING_PREFECT_UI_ENABLED": "true",
    "INFRAHUB_TESTING_DOCKER_PULL": "true",
    "INFRAHUB_TESTING_SCHEMA_STRICT_MODE": "true",
}


@dataclass
class InfrahubDockerCompose(DockerCompose):
    project_name: str | None = None
    env_vars: dict[str, str] = field(default_factory=dict)
    deployment_type: str | None = None

    @classmethod
    def init(
        cls, directory: Path | None = None, version: str | None = None, deployment_type: str | None = None
    ) -> Self:
        if not directory:
            directory = Path.cwd()

        if not version:
            version = infrahub_version

        infrahub_image_version = os.environ.get("INFRAHUB_TESTING_IMAGE_VER", None)
        if version == "local" and infrahub_image_version:
            version = infrahub_image_version

        compose = cls(project_name=cls.generate_project_name(), context=directory, deployment_type=deployment_type)
        compose.create_docker_file(directory=directory)
        compose.create_env_file(directory=directory, version=version)

        return compose

    def get_env_var(self, key: str) -> str:
        if not self.env_vars:
            raise ValueError("env_vars hasn't been initialized yet")
        if key not in self.env_vars:
            raise ValueError(f"{key} is not set in the environment variables")
        return self.env_vars[key]

    @property
    def use_neo4j_enterprise(self) -> bool:
        return "enterprise" in self.get_env_var("NEO4J_DOCKER_IMAGE")

    @property
    def internal_backup_dir(self) -> Path:
        return Path(self.get_env_var("INFRAHUB_TESTING_INTERNAL_DB_BACKUP_DIRECTORY"))

    @property
    def external_backup_dir(self) -> Path:
        return Path(self.context) / Path(self.get_env_var("INFRAHUB_TESTING_LOCAL_DB_BACKUP_DIRECTORY"))

    @classmethod
    def generate_project_name(cls) -> str:
        project_id = str(uuid.uuid4())[:8]
        return f"infrahub-test-{project_id}"

    def create_docker_file(self, directory: Path) -> Path:
        current_directory = Path(__file__).resolve().parent
        compose_file_name = (
            "docker-compose-cluster.test.yml" if self.deployment_type == "cluster" else "docker-compose.test.yml"
        )
        compose_file = current_directory / compose_file_name

        test_compose_file = directory / "docker-compose.yml"
        test_compose_file.write_bytes(compose_file.read_bytes())

        for file in ["haproxy.cfg", "prometheus.yml"]:
            config_file = current_directory / file

            test_config_file = directory / file
            test_config_file.write_bytes(config_file.read_bytes())

        return test_compose_file

    def create_env_file(self, directory: Path, version: str) -> Path:
        env_file = directory / ".env"

        PROJECT_ENV_VARIABLES.update({"INFRAHUB_TESTING_IMAGE_VERSION": version})
        if os.environ.get("INFRAHUB_TESTING_ENTERPRISE"):
            PROJECT_ENV_VARIABLES.update(
                {
                    "INFRAHUB_TESTING_DOCKER_IMAGE": "registry.opsmill.io/opsmill/infrahub-enterprise",
                    "INFRAHUB_TESTING_DOCKER_ENTRYPOINT": f"gunicorn --config community/backend/infrahub/serve/gunicorn_config.py -w {os.environ.get('INFRAHUB_TESTING_WEB_CONCURRENCY', 4)} --logger-class infrahub.serve.log.GunicornLogger infrahub_enterprise.server:app",  # noqa: E501
                    "INFRAHUB_TESTING_WORKFLOW_DEFAULT_WORKER_TYPE": "infrahubentasync",
                    "INFRAHUB_TESTING_PREFECT_UI_ENABLED": "false",
                    "NEO4J_DOCKER_IMAGE": "neo4j:2025.03.0-enterprise",
                }
            )

        with env_file.open(mode="w", encoding="utf-8") as file:
            for key, value in PROJECT_ENV_VARIABLES.items():
                env_var_value = os.environ.get(key, value)
                file.write(f"{key}={env_var_value}\n")
                self.env_vars[key] = env_var_value

        return env_file.absolute()

    def restart(self) -> None:
        """
        Restart the docker compose environment.

        TODO Would be good to contribute this upstream
        """
        cmd = self.compose_command_property[:]
        cmd += ["restart"]

        if self.services:
            cmd.extend(self.services)
        self._run_command(cmd=cmd)

    def start_container(self, service_name: str | list[str]) -> None:
        """
        Starts a specific service of the docker compose environment.

        TODO Would be good to contribute this upstream
        """
        base_cmd = self.compose_command_property or []

        # pull means running a separate command before starting
        if self.pull:
            pull_cmd = [*base_cmd, "pull"]
            if isinstance(service_name, list):
                pull_cmd.extend(service_name)
            else:
                pull_cmd.append(service_name)
            self._run_command(cmd=pull_cmd)

        up_cmd = [*base_cmd, "up"]

        if self.get_env_var("INFRAHUB_TESTING_DOCKER_PULL") == "false":
            up_cmd.extend(["--pull", "never"])

        # build means modifying the up command
        if self.wait:
            up_cmd.append("--wait")
        else:
            # we run in detached mode instead of blocking
            up_cmd.append("--detach")

        if isinstance(service_name, list):
            up_cmd.extend(service_name)
        else:
            up_cmd.append(service_name)
        self._run_command(cmd=up_cmd)

    # TODO would be good to the support for project_name upstream
    @cached_property
    def compose_command_property(self) -> list[str]:
        docker_compose_cmd = [self.docker_command_path or "docker", "compose"]
        if self.compose_file_name:
            for file in self.compose_file_name:
                docker_compose_cmd += ["-f", file]
        if self.project_name:
            docker_compose_cmd += ["--project-name", self.project_name]
        if self.env_file:
            docker_compose_cmd += ["--env-file", self.env_file]
        return docker_compose_cmd

    def get_services_port(self) -> dict[str, int]:
        return {
            service_name: int(self.get_service_port(service_name=service_data.container, port=service_data.port) or 0)
            for service_name, service_data in INFRAHUB_SERVICES.items()
        }

    def database_create_backup(self, backup_name: str = "neo4j_database.backup", dest_dir: Path | None = None) -> None:
        assert self.use_neo4j_enterprise

        self.exec_in_container(
            command=[
                "neo4j-admin",
                "database",
                "backup",
                "--compress=false",
                "--to-path",
                str(self.internal_backup_dir),
            ],
            service_name="database",
        )

        if dest_dir:
            backup_files = list(self.external_backup_dir.glob("*.backup"))
            if not backup_files:
                raise FileNotFoundError(f"No .backup files found in {self.external_backup_dir}")

            backup_file = backup_files[0]
            shutil.copy(
                backup_file,
                dest_dir / backup_name,
            )

    def database_restore_backup(self, backup_file: Path) -> None:  # noqa: PLR0915
        assert self.use_neo4j_enterprise

        shutil.copy(
            str(backup_file),
            str(self.external_backup_dir / backup_file.name),
        )
        service_name = "database"

        if self.deployment_type != "cluster":  # noqa: PLR1702
            try:
                self.get_container(service_name=service_name)
            except ContainerIsNotRunning:
                self.start_container(service_name=service_name)

            self.exec_in_container(
                command=["cypher-shell", "-u", "neo4j", "-p", "admin", "STOP DATABASE neo4j;"],
                service_name=service_name,
            )

            self.exec_in_container(
                command=[
                    "neo4j-admin",
                    "database",
                    "restore",
                    "--overwrite-destination",
                    "--from-path",
                    str(self.internal_backup_dir / backup_file.name),
                ],
                service_name=service_name,
            )

            self.exec_in_container(
                command=["chown", "-R", "neo4j:neo4j", "/data"],
                service_name=service_name,
            )

            (restore_output, _, _) = self.exec_in_container(
                command=[
                    "cypher-shell",
                    "--format",
                    "plain",
                    "-d",
                    "system",
                    "-u",
                    "neo4j",
                    "-p",
                    "admin",
                    "START DATABASE neo4j;",
                ],
                service_name=service_name,
            )

            for _ in range(3):
                (stdout, _, _) = self.exec_in_container(
                    command=[
                        "cypher-shell",
                        "--format",
                        "plain",
                        "-d",
                        "system",
                        "-u",
                        "neo4j",
                        "-p",
                        "admin",
                        "SHOW DATABASES WHERE name = 'neo4j' AND currentStatus = 'online';",
                    ],
                    service_name=service_name,
                )
                if stdout:
                    break
                time.sleep(5)
            else:
                (debug_logs, _, _) = self.exec_in_container(
                    command=["cat", "logs/debug.log"],
                    service_name=service_name,
                )
                raise Exception(f"Failed to restore database:\n{restore_output}\nDebug logs:\n{debug_logs}")

            old_services = self.services
            self.services = ["infrahub-server", "task-worker"]
            self.stop(down=False)
            try:
                self.start()
            except Exception as exc:
                stdout, stderr = self.get_logs()
                raise Exception(f"Failed to start docker compose:\nStdout:\n{stdout}\nStderr:\n{stderr}") from exc
            self.services = old_services
        else:
            print("Cluster mode detected")
            try:
                self.get_container(service_name=service_name)
                self.get_container(service_name="database-core2")
                self.get_container(service_name="database-core3")
            except ContainerIsNotRunning:
                self.start_container("database", "database-core2", "database-core3")

            # Waiting for cluster to stabilize...
            time.sleep(10)

            self.exec_in_container(
                command=["cypher-shell", "-u", "neo4j", "-p", "admin", "DROP DATABASE neo4j;"],
                service_name=service_name,
            )

            self.exec_in_container(
                command=["rm", "-rf", "/data/databases/neo4j"],
                service_name=service_name,
            )
            self.exec_in_container(
                command=["rm", "-rf", "/data/transactions/neo4j"],
                service_name=service_name,
            )

            self.exec_in_container(
                command=[
                    "neo4j-admin",
                    "database",
                    "restore",
                    "--from-path",
                    str(self.internal_backup_dir / backup_file.name),
                    "neo4j",
                ],
                service_name=service_name,
            )

            cmd = self.compose_command_property[:]
            cmd += ["restart", "database"]
            self._run_command(cmd=cmd)

            main_node = service_name
            cluster_nodes = ["database", "database-core2", "database-core3"]

            for attempt in range(3):
                try:
                    (stdout, _, _) = self.exec_in_container(
                        command=[
                            "cypher-shell",
                            "--format",
                            "plain",
                            "-d",
                            "system",
                            "-u",
                            "neo4j",
                            "-p",
                            "admin",
                            "SHOW DATABASES YIELD name, address, currentStatus WHERE name = 'system' RETURN address, currentStatus",
                        ],
                        service_name=main_node,
                    )
                except Exception:
                    time.sleep(10)
                    continue

                raw_output = stdout
                nodes_status = dict.fromkeys(cluster_nodes, False)
                online_count = 0
                total_entries = 0

                try:
                    for line_raw in stdout.splitlines():
                        line = line_raw.strip()
                        if not line or line.startswith("address"):
                            continue

                        total_entries += 1
                        if "online" in line:
                            online_count += 1
                            for node in cluster_nodes:
                                node_pattern = f'"{node}:'
                                if node_pattern in line:
                                    nodes_status[node] = True
                                    break
                    if all(nodes_status.values()) and online_count == len(cluster_nodes):
                        break
                except Exception as e:
                    print(f"Error parsing database status on attempt {attempt + 1}: {e}")

                print(f"Waiting for all nodes to be online. Current status: {nodes_status}")
                time.sleep(5)
            else:
                debug_logs = {}
                for node in cluster_nodes:
                    try:
                        (logs, _, _) = self.exec_in_container(
                            command=["cat", "logs/debug.log"],
                            service_name=node,
                        )
                        debug_logs[node] = logs
                    except Exception as e:
                        debug_logs[node] = f"Could not retrieve logs: {str(e)}"

                debug_info = f"Raw output from SHOW DATABASES command:\n{raw_output}\n\n"
                debug_info += f"Final node status: {nodes_status}\n\n"

                status_str = ", ".join(
                    [f"{node}: {'online' if status else 'offline'}" for node, status in nodes_status.items()]
                )
                logs_str = debug_info + "\n\n".join(
                    [f"--- {node} logs ---\n{logs}" for node, logs in debug_logs.items()]
                )

                raise Exception(
                    f"Failed to restore database cluster. Node status: {status_str}\nDebug logs:\n{logs_str}"
                )

            server_id = None
            try:
                stdout, _, _ = self.exec_in_container(
                    command=[
                        "cypher-shell",
                        "--format",
                        "plain",
                        "-d",
                        "system",
                        "-u",
                        "neo4j",
                        "-p",
                        "admin",
                        'SHOW SERVERS YIELD name, address WHERE address = "database:7687" RETURN name;',
                    ],
                    service_name=service_name,
                )

                lines = stdout.splitlines()
                for line_raw in lines:
                    line = line_raw.strip()
                    if not line or line == "name" or line.startswith("+"):
                        continue
                    server_id = line.strip('"')
                    break
            except Exception as e:
                print(f"Error retrieving server ID with direct query: {e}")

            if server_id:
                self.exec_in_container(
                    command=[
                        "cypher-shell",
                        "-d",
                        "system",
                        "-u",
                        "neo4j",
                        "-p",
                        "admin",
                        f"CREATE DATABASE neo4j TOPOLOGY 3 PRIMARIES OPTIONS {{ existingData: 'use', existingDataSeedInstance: '{server_id}' }};",
                    ],
                    service_name=service_name,
                )
            self.start()
            print("Database restored successfully")



import importlib
import logging
import os
from typing import TYPE_CHECKING, Any

import structlog
from pydantic import TypeAdapter
from structlog.dev import plain_traceback

if TYPE_CHECKING:
    from structlog.types import Processor

INFRAHUB_PRODUCTION = TypeAdapter(bool).validate_python(os.environ.get("INFRAHUB_PRODUCTION", True))
INFRAHUB_LOG_LEVEL = os.environ.get("INFRAHUB_LOG_LEVEL", "INFO")


def clear_log_context() -> None:
    structlog.contextvars.clear_contextvars()


def get_logger(name: str = "infrahub") -> structlog.stdlib.BoundLogger:
    return structlog.stdlib.get_logger(name)


def get_run_logger(name: str = "infrahub.tasks") -> logging.Logger:
    return logging.getLogger(name)


def get_log_data() -> dict[str, Any]:
    return structlog.contextvars.get_contextvars()


def set_log_data(key: str, value: Any) -> None:
    structlog.contextvars.bind_contextvars(**{key: value})


def configure_logging(production: bool, log_level: str) -> None:
    # Importing prefect.main here triggers prefect.logging.configuration.setup_logging()
    # to be executed, this function wipes out the previous logging configuration and
    # starts from a clean slate. After this has been imported once we can reinject
    # the infrahub logger
    importlib.import_module("prefect.main")
    shared_processors: list[Processor] = [
        structlog.contextvars.merge_contextvars,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
    ]
    logging.getLogger("httpx").setLevel(logging.ERROR)

    if production:
        shared_processors.append(structlog.processors.format_exc_info)

    structlog.configure(
        processors=shared_processors + [structlog.stdlib.ProcessorFormatter.wrap_for_formatter],
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    log_renderer: structlog.types.Processor
    if production:
        log_renderer = structlog.processors.JSONRenderer()
    else:
        log_renderer = structlog.dev.ConsoleRenderer(exception_formatter=plain_traceback)

    formatter = structlog.stdlib.ProcessorFormatter(
        foreign_pre_chain=shared_processors,
        processors=[structlog.stdlib.ProcessorFormatter.remove_processors_meta, log_renderer],
    )

    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    for existing_handler in root_logger.handlers:
        if isinstance(existing_handler, logging.StreamHandler):
            root_logger.removeHandler(existing_handler)

    root_logger.addHandler(handler)
    root_logger.setLevel(log_level)


configure_logging(production=INFRAHUB_PRODUCTION, log_level=INFRAHUB_LOG_LEVEL)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.core.branch import Branch
    from infrahub.database import InfrahubDatabase
    from infrahub.permissions.types import AssignedPermissions


class PermissionBackend(ABC):
    @abstractmethod
    async def load_permissions(
        self, db: InfrahubDatabase, branch: Branch, account_session: AccountSession
    ) -> AssignedPermissions: ...



from typing import TYPE_CHECKING, Any

from graphql import GraphQLResolveInfo
from infrahub_sdk.utils import deep_merge_dict, extract_fields

from infrahub.core.branch.models import Branch
from infrahub.core.constants import BranchSupportType, RelationshipHierarchyDirection
from infrahub.core.manager import NodeManager
from infrahub.core.query.node import NodeGetHierarchyQuery
from infrahub.core.schema.node_schema import NodeSchema
from infrahub.core.schema.relationship_schema import RelationshipSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase

from ..loaders.peers import PeerRelationshipsDataLoader, QueryPeerParams
from ..types import RELATIONS_PROPERTY_MAP, RELATIONS_PROPERTY_MAP_REVERSED

if TYPE_CHECKING:
    from infrahub.core.schema import MainSchemaTypes

    from ..initialization import GraphqlContext


class ManyRelationshipResolver:
    def __init__(self) -> None:
        self._data_loader_instances: dict[QueryPeerParams, PeerRelationshipsDataLoader] = {}

    async def get_descendant_ids(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | None,
        parent_id: str,
        node_schema: NodeSchema,
    ) -> list[str]:
        async with db.start_session(read_only=True) as dbs:
            query = await NodeGetHierarchyQuery.init(
                db=dbs,
                direction=RelationshipHierarchyDirection.DESCENDANTS,
                node_id=parent_id,
                node_schema=node_schema,
                at=at,
                branch=branch,
            )
            await query.execute(db=dbs)
        return list(query.get_peer_ids())

    async def get_peer_count(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        at: Timestamp | None,
        ids: list[str],
        source_kind: str,
        rel_schema: RelationshipSchema,
        filters: dict[str, Any],
    ) -> int:
        async with db.start_session(read_only=True) as dbs:
            return await NodeManager.count_peers(
                db=dbs,
                ids=ids,
                source_kind=source_kind,
                schema=rel_schema,
                filters=filters,
                at=at,
                branch=branch,
                branch_agnostic=rel_schema.branch is BranchSupportType.AGNOSTIC,
            )

    async def resolve(
        self,
        parent: dict,
        info: GraphQLResolveInfo,
        include_descendants: bool = False,
        offset: int | None = None,
        limit: int | None = None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Resolver for relationships of cardinality=one for Edged responses

        This resolver is used for paginated responses and as such we redefined the requested
        fields by only reusing information below the 'node' key.
        """
        # Extract the InfraHub schema by inspecting the GQL Schema

        node_schema: MainSchemaTypes = info.parent_type.graphene_type._meta.schema  # type: ignore[attr-defined]

        graphql_context: GraphqlContext = info.context

        # Extract the name of the fields in the GQL query
        fields = await extract_fields(info.field_nodes[0].selection_set)
        edges = fields.get("edges", {})
        node_fields = edges.get("node", {})
        property_fields = edges.get("properties", {})
        for key, value in property_fields.items():
            mapped_name = RELATIONS_PROPERTY_MAP[key]
            node_fields[mapped_name] = value

        filters = {
            f"{info.field_name}__{key}": value
            for key, value in kwargs.items()
            if ("__" in key and value) or key in ["id", "ids"]
        }

        response: dict[str, Any] = {"edges": [], "count": None}

        # Extract the schema of the node on the other end of the relationship from the GQL Schema
        node_rel = node_schema.get_relationship(info.field_name)
        source_kind = node_schema.kind
        ids = [parent["id"]]
        if isinstance(node_schema, NodeSchema):
            if node_schema.hierarchy:
                source_kind = node_schema.hierarchy

            if include_descendants:
                descendant_ids = await self.get_descendant_ids(
                    db=graphql_context.db,
                    branch=graphql_context.branch,
                    at=graphql_context.at,
                    parent_id=ids[0],
                    node_schema=node_schema,
                )
                ids.extend(descendant_ids)

        if "count" in fields:
            peer_count = await self.get_peer_count(
                db=graphql_context.db,
                branch=graphql_context.branch,
                at=graphql_context.at,
                ids=ids,
                source_kind=source_kind,
                rel_schema=node_rel,
                filters=filters,
            )
            response["count"] = peer_count

        if not node_fields:
            return response

        if offset or limit:
            node_graph = await self._get_entities_simple(
                db=graphql_context.db,
                branch=graphql_context.branch,
                ids=ids,
                at=graphql_context.at,
                related_node_ids=graphql_context.related_node_ids,
                source_kind=source_kind,
                rel_schema=node_rel,
                filters=filters,
                node_fields=node_fields,
                offset=offset,
                limit=limit,
            )
        else:
            node_graph = await self._get_entities_with_data_loader(
                db=graphql_context.db,
                branch=graphql_context.branch,
                ids=ids,
                at=graphql_context.at,
                related_node_ids=graphql_context.related_node_ids,
                source_kind=source_kind,
                rel_schema=node_rel,
                filters=filters,
                node_fields=node_fields,
            )

        if not node_graph:
            return response

        entries = []
        for node in node_graph:
            entry: dict[str, dict[str, Any]] = {"node": {}, "properties": {}}
            for key, mapped in RELATIONS_PROPERTY_MAP_REVERSED.items():
                value = node.pop(key, None)
                if value:
                    entry["properties"][mapped] = value
            entry["node"] = node
            entries.append(entry)

        response["edges"] = entries
        return response

    async def _get_entities_simple(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        ids: list[str],
        at: Timestamp | None,
        related_node_ids: set[str] | None,
        source_kind: str,
        rel_schema: RelationshipSchema,
        filters: dict[str, Any],
        node_fields: dict[str, Any],
        offset: int | None = None,
        limit: int | None = None,
    ) -> list[dict[str, Any]] | None:
        async with db.start_session(read_only=True) as dbs:
            objs = await NodeManager.query_peers(
                db=dbs,
                ids=ids,
                source_kind=source_kind,
                schema=rel_schema,
                filters=filters,
                fields=node_fields,
                offset=offset,
                limit=limit,
                at=at,
                branch=branch,
                branch_agnostic=rel_schema.branch is BranchSupportType.AGNOSTIC,
                fetch_peers=True,
            )
            if not objs:
                return None
            return [await obj.to_graphql(db=dbs, fields=node_fields, related_node_ids=related_node_ids) for obj in objs]

    async def _get_entities_with_data_loader(
        self,
        db: InfrahubDatabase,
        branch: Branch,
        ids: list[str],
        at: Timestamp | None,
        related_node_ids: set[str] | None,
        source_kind: str,
        rel_schema: RelationshipSchema,
        filters: dict[str, Any],
        node_fields: dict[str, Any],
    ) -> list[dict[str, Any]] | None:
        if node_fields and "display_label" in node_fields:
            schema_branch = db.schema.get_schema_branch(name=branch.name)
            display_label_fields = schema_branch.generate_fields_for_display_label(name=rel_schema.peer)
            if display_label_fields:
                node_fields = deep_merge_dict(dicta=node_fields, dictb=display_label_fields)

        if node_fields and "hfid" in node_fields:
            peer_schema = db.schema.get(name=rel_schema.peer, branch=branch, duplicate=False)
            hfid_fields = peer_schema.generate_fields_for_hfid()
            if hfid_fields:
                node_fields = deep_merge_dict(dicta=node_fields, dictb=hfid_fields)

        query_params = QueryPeerParams(
            branch=branch,
            source_kind=source_kind,
            schema=rel_schema,
            filters=filters,
            fields=node_fields,
            at=at,
            branch_agnostic=rel_schema.branch is BranchSupportType.AGNOSTIC,
        )
        if query_params in self._data_loader_instances:
            loader = self._data_loader_instances[query_params]
        else:
            loader = PeerRelationshipsDataLoader(db=db, query_params=query_params)
            self._data_loader_instances[query_params] = loader
        all_peer_rels = []
        for node_id in ids:
            node_peer_rels = await loader.load(key=node_id)
            all_peer_rels.extend(node_peer_rels)
        if not all_peer_rels:
            return None
        async with db.start_session(read_only=True) as dbs:
            return [
                await obj.to_graphql(db=dbs, fields=node_fields, related_node_ids=related_node_ids)
                for obj in all_peer_rels
            ]



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from infrahub import config
from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.constants import SchemaPathType
from infrahub.core.initialization import initialization
from infrahub.core.migrations.shared import InternalSchemaMigration, MigrationResult, SchemaMigration
from infrahub.core.path import SchemaPath
from infrahub.core.schema import GenericSchema, NodeSchema
from infrahub.core.schema.attribute_parameters import NumberAttributeParameters
from infrahub.core.validators.attribute.min_max import AttributeNumberChecker
from infrahub.core.validators.enum import ConstraintIdentifier
from infrahub.core.validators.model import SchemaConstraintValidatorRequest
from infrahub.lock import initialize_lock
from infrahub.log import get_logger
from infrahub.types import Number

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class Migration031(InternalSchemaMigration):
    """
    Some nodes with invalid number attributes may have been created as min/max/excluded_values were not working properly.
    This migration indicates corrupted nodes. If strict mode is disabled, both this migration and min/max/excludes_values constraints are disabled,
    so that users can carry one with their corrupted data without any failure.
    """

    name: str = "031_check_number_attributes"
    minimum_version: int = 30
    migrations: Sequence[SchemaMigration] = []

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        """Retrieve all number attributes that have a min/max/excluded_values
        For any of these attributes, check if corresponding existing nodes are valid."""

        if not config.SETTINGS.main.schema_strict_mode:
            return MigrationResult()

        # load schemas from database into registry
        initialize_lock()
        await initialization(db=db)

        node_id_to_error_message = {}

        branches = await Branch.get_list(db=db)
        for branch in branches:  # noqa
            schema_branch = await registry.schema.load_schema_from_db(db=db, branch=branch)
            for node_schema_kind in schema_branch.node_names:
                schema = schema_branch.get_node(name=node_schema_kind, duplicate=False)
                if not isinstance(schema, (NodeSchema, GenericSchema)):
                    continue

                for attr in schema.attributes:
                    if attr.kind != Number.label:
                        continue

                    # Check if the attribute has a min/max/excluded_values being violated
                    if isinstance(attr.parameters, NumberAttributeParameters) and (
                        attr.parameters.min_value is not None
                        or attr.parameters.max_value is not None
                        or attr.parameters.excluded_values
                    ):
                        request = SchemaConstraintValidatorRequest(
                            branch=branch,
                            constraint_name=ConstraintIdentifier.ATTRIBUTE_PARAMETERS_MIN_VALUE_UPDATE.value,
                            node_schema=schema,
                            schema_path=SchemaPath(
                                path_type=SchemaPathType.ATTRIBUTE, schema_kind=schema.kind, field_name=attr.name
                            ),
                            schema_branch=db.schema.get_schema_branch(name=registry.default_branch),
                        )

                        constraint_checker = AttributeNumberChecker(db=db, branch=branch)
                        grouped_data_paths = await constraint_checker.check(request)
                        data_paths = grouped_data_paths[0].get_all_data_paths()
                        for data_path in data_paths:
                            # Avoid having duplicated error messages for nodes present on multiple branches.
                            if data_path.node_id not in node_id_to_error_message:
                                node_id_to_error_message[data_path.node_id] = (
                                    f"Node {data_path.node_id} on branch {branch.name} "
                                    f"has an invalid Number attribute {data_path.field_name}: {data_path.value}"
                                )

        if len(node_id_to_error_message) == 0:
            return MigrationResult()

        error_str = (
            "Following nodes attributes values must be updated to not violate corresponding min_value, "
            "max_value or excluded_values schema constraints"
        )
        errors_messages = list(node_id_to_error_message.values())
        return MigrationResult(errors=[error_str] + errors_messages)

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        result = MigrationResult()
        return result



from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from infrahub.core.migrations.shared import MigrationResult
from infrahub.log import get_logger

from ..shared import InternalSchemaMigration, SchemaMigration
from .m018_uniqueness_nulls import validate_nulls_in_uniqueness_constraints

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class Migration025(InternalSchemaMigration):
    name: str = "025_validate_nulls_in_uniqueness_constraints"
    minimum_version: int = 24
    migrations: Sequence[SchemaMigration] = []

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        return await validate_nulls_in_uniqueness_constraints(db=db)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from fastapi import APIRouter, Body, Depends, Path, Query, Request
from graphql import graphql
from pydantic import BaseModel, Field

from infrahub.api.dependencies import BranchParams, get_branch_params, get_current_user, get_db
from infrahub.context import InfrahubContext
from infrahub.core import registry
from infrahub.core.constants import InfrahubKind
from infrahub.core.protocols import CoreGraphQLQuery
from infrahub.database import InfrahubDatabase  # noqa: TC001
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.api.dependencies import build_graphql_query_permission_checker
from infrahub.graphql.initialization import prepare_graphql_params
from infrahub.graphql.metrics import (
    GRAPHQL_DURATION_METRICS,
    GRAPHQL_QUERY_DEPTH_METRICS,
    GRAPHQL_QUERY_HEIGHT_METRICS,
    GRAPHQL_QUERY_OBJECTS_METRICS,
    GRAPHQL_QUERY_VARS_METRICS,
    GRAPHQL_RESPONSE_SIZE_METRICS,
    GRAPHQL_TOP_LEVEL_QUERIES_METRICS,
)
from infrahub.graphql.utils import extract_data
from infrahub.groups.models import RequestGraphQLQueryGroupUpdate
from infrahub.log import get_logger
from infrahub.workflows.catalogue import GRAPHQL_QUERY_GROUP_UPDATE

if TYPE_CHECKING:
    from infrahub.auth import AccountSession
    from infrahub.graphql.auth.query_permission_checker.checker import GraphQLQueryPermissionChecker
    from infrahub.services import InfrahubServices


log = get_logger()
router = APIRouter(prefix="/query")


class QueryPayload(BaseModel):
    variables: dict[str, str] = Field(default_factory=dict)


async def execute_query(
    db: InfrahubDatabase,
    request: Request,
    branch_params: BranchParams,
    query_id: str,
    params: dict[str, str],
    update_group: bool,
    subscribers: list[str],
    permission_checker: GraphQLQueryPermissionChecker,
    account_session: AccountSession,
) -> dict[str, Any]:
    gql_query = await registry.manager.get_one_by_id_or_default_filter(
        db=db, id=query_id, kind=CoreGraphQLQuery, branch=branch_params.branch, at=branch_params.at
    )

    context = InfrahubContext.init(branch=branch_params.branch, account=account_session)

    gql_params = await prepare_graphql_params(
        db=db,
        branch=branch_params.branch,
        at=branch_params.at,
        account_session=account_session,
        service=request.app.state.service,
    )
    schema_branch = db.schema.get_schema_branch(name=branch_params.branch.name)

    analyzed_query = InfrahubGraphQLQueryAnalyzer(
        query=gql_query.query.value,
        schema=gql_params.schema,
        schema_branch=schema_branch,
        branch=branch_params.branch,
    )
    await permission_checker.check(
        db=db,
        account_session=account_session,
        analyzed_query=analyzed_query,
        query_parameters=gql_params,
        branch=branch_params.branch,
    )

    labels = {
        "type": "mutation" if analyzed_query.contains_mutation else "query",
        "branch": branch_params.branch.name,
        "operation": "",
        "name": gql_query.name.value,
        "query_id": query_id,
    }

    with GRAPHQL_DURATION_METRICS.labels(**labels).time():
        result = await graphql(
            schema=gql_params.schema,
            source=gql_query.query.value,
            context_value=gql_params.context,
            root_value=None,
            variable_values=params,
        )

    data = extract_data(query_name=gql_query.name.value, result=result)

    GRAPHQL_RESPONSE_SIZE_METRICS.labels(**labels).observe(len(data))
    GRAPHQL_QUERY_DEPTH_METRICS.labels(**labels).observe(await analyzed_query.calculate_depth())
    GRAPHQL_QUERY_HEIGHT_METRICS.labels(**labels).observe(await analyzed_query.calculate_height())
    GRAPHQL_QUERY_VARS_METRICS.labels(**labels).observe(len(analyzed_query.variables))
    GRAPHQL_TOP_LEVEL_QUERIES_METRICS.labels(**labels).observe(analyzed_query.nbr_queries)
    GRAPHQL_QUERY_OBJECTS_METRICS.labels(**labels).observe(len(analyzed_query.query_report.impacted_models))

    response_payload: dict[str, Any] = {"data": data}

    related_node_ids = gql_params.context.related_node_ids or set()

    if update_group:
        service: InfrahubServices = request.app.state.service
        model = RequestGraphQLQueryGroupUpdate(
            branch=branch_params.branch.name,
            query_id=gql_query.id,
            query_name=gql_query.name.value,
            related_node_ids=sorted(related_node_ids),
            subscribers=sorted(subscribers),
            params=params,
        )
        await service.workflow.submit_workflow(
            workflow=GRAPHQL_QUERY_GROUP_UPDATE, context=context, parameters={"model": model}
        )

    return response_payload


@router.post("/{query_id}")
async def graphql_query_post(
    request: Request,
    payload: QueryPayload = Body(
        QueryPayload(),  # noqa: B008
        description="Payload of the request, must be used to provide the variables",
    ),
    query_id: str = Path(description="ID or Name of the GraphQL query to execute"),
    subscribers: list[str] = Query(
        [], description=f"List of subscribers to attach to the {InfrahubKind.GRAPHQLQUERYGROUP}"
    ),
    update_group: bool = Query(
        False,
        description=f"When True create or update a {InfrahubKind.GRAPHQLQUERYGROUP} with all nodes related to this query.",
    ),
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    account_session: AccountSession = Depends(get_current_user),
    permission_checker: GraphQLQueryPermissionChecker = Depends(build_graphql_query_permission_checker),
) -> dict:
    return await execute_query(
        db=db,
        request=request,
        branch_params=branch_params,
        query_id=query_id,
        params=payload.variables,
        update_group=update_group,
        subscribers=subscribers,
        permission_checker=permission_checker,
        account_session=account_session,
    )


@router.get("/{query_id}")
async def graphql_query_get(
    request: Request,
    query_id: str = Path(description="ID or Name of the GraphQL query to execute"),
    subscribers: list[str] = Query(
        [], description=f"List of subscribers to attach to the {InfrahubKind.GRAPHQLQUERYGROUP}"
    ),
    update_group: bool = Query(
        False,
        description=f"When True create or update a {InfrahubKind.GRAPHQLQUERYGROUP} with all nodes related to this query.",
    ),
    db: InfrahubDatabase = Depends(get_db),
    branch_params: BranchParams = Depends(get_branch_params),
    account_session: AccountSession = Depends(get_current_user),
    permission_checker: GraphQLQueryPermissionChecker = Depends(build_graphql_query_permission_checker),
) -> dict:
    params = {
        key: value
        for key, value in request.query_params.items()
        if key not in ["branch", "at", "update_group", "subscribers"]
    }

    return await execute_query(
        db=db,
        request=request,
        branch_params=branch_params,
        query_id=query_id,
        params=params,
        update_group=update_group,
        subscribers=subscribers,
        permission_checker=permission_checker,
        account_session=account_session,
    )



from infrahub.core.constants import BranchSupportType

from ...attribute_schema import AttributeSchema as Attr
from ...node_schema import NodeSchema

builtin_tag = NodeSchema(
    name="Tag",
    namespace="Builtin",
    description="Standard Tag object to attach to other objects to provide some context.",
    include_in_menu=True,
    icon="mdi:tag-multiple",
    label="Tag",
    default_filter="name__value",
    order_by=["name__value"],
    display_labels=["name__value"],
    branch=BranchSupportType.AWARE,
    uniqueness_constraints=[["name__value"]],
    attributes=[
        Attr(name="name", kind="Text", unique=True),
        Attr(name="description", kind="Text", optional=True),
    ],
)



from pydantic import BaseModel

from infrahub.core.constants import RelationshipCardinality
from infrahub.core.schema import NodeSchema


class SchemaMappingValue(BaseModel):
    is_mandatory: bool
    source_field_name: str | None = None  # None means there is no corresponding source field name
    relationship_cardinality: RelationshipCardinality | None = None


SchemaMapping = dict[str, SchemaMappingValue]


def get_schema_mapping(source_schema: NodeSchema, target_schema: NodeSchema) -> SchemaMapping:
    """
    Return fields mapping meant to be used for converting a node from `source_kind` to `target_kind`.
    For any field of the target kind, field of the source kind will be matched if:
    - It's an attribute with identical name and type.
    - It's a relationship with identical peer kind and cardinality.
    If there is no match, the mapping will only indicate whether the field is mandatory or not.
    """

    target_field_to_source_field = {}

    # Create lookup dictionaries for source attributes and relationships
    source_attrs = {attr.name: attr for attr in source_schema.attributes}
    source_rels = {rel.name: rel for rel in source_schema.relationships}

    # Process attributes
    for target_attr in target_schema.attributes:
        source_attr = source_attrs.get(target_attr.name)
        if source_attr and source_attr.kind == target_attr.kind:
            target_field_to_source_field[target_attr.name] = SchemaMappingValue(
                source_field_name=source_attr.name, is_mandatory=not target_attr.optional
            )
        else:
            target_field_to_source_field[target_attr.name] = SchemaMappingValue(is_mandatory=not target_attr.optional)

    # Process relationships
    for target_rel in target_schema.relationships:
        source_rel = source_rels.get(target_rel.name)
        if source_rel and source_rel.peer == target_rel.peer and source_rel.cardinality == target_rel.cardinality:
            target_field_to_source_field[target_rel.name] = SchemaMappingValue(
                source_field_name=source_rel.name,
                is_mandatory=not target_rel.optional,
                relationship_cardinality=target_rel.cardinality,
            )
        else:
            target_field_to_source_field[target_rel.name] = SchemaMappingValue(
                is_mandatory=not target_rel.optional,
                relationship_cardinality=target_rel.cardinality,
            )

    return target_field_to_source_field



from typing import Any

from pydantic import BaseModel

from infrahub.core.attribute import BaseAttribute
from infrahub.core.branch import Branch
from infrahub.core.constants import RelationshipCardinality
from infrahub.core.manager import NodeManager
from infrahub.core.node import Node
from infrahub.core.node.create import create_node
from infrahub.core.query.relationship import GetAllPeersIds
from infrahub.core.query.resource_manager import PoolChangeReserved
from infrahub.core.relationship import RelationshipManager
from infrahub.core.schema import NodeSchema
from infrahub.database import InfrahubDatabase


class InputDataForDestField(BaseModel):  # Only one of these fields can be not None
    attribute_value: Any | None = None
    peer_id: str | None = None
    peers_ids: list[str] | None = None

    @property
    def value(self) -> Any:
        fields = [self.attribute_value, self.peer_id, self.peers_ids]
        set_fields = [f for f in fields if f is not None]
        if len(set_fields) != 1:
            raise ValueError("Exactly one of attribute_value, peer_id, or peers_ids must be set")
        return set_fields[0]


class InputForDestField(BaseModel):  # Only one of these fields can be not None
    source_field: str | None = None
    data: InputDataForDestField | None = None

    @property
    def value(self) -> Any:
        if self.source_field is not None and self.data is not None:
            raise ValueError("Only one of source_field or data can be set")
        if self.source_field is None and self.data is None:
            raise ValueError("Either source_field or data must be set")
        return self.source_field if self.source_field is not None else self.data


async def get_out_rels_peers_ids(node: Node, db: InfrahubDatabase) -> list[str]:
    all_peers: list[Node] = []
    for name in node._relationships:
        relm: RelationshipManager = getattr(node, name)
        peers = await relm.get_peers(db=db)
        all_peers.extend(peers.values())
    return [peer.id for peer in all_peers]


async def build_data_new_node(db: InfrahubDatabase, mapping: dict[str, InputForDestField], node: Node) -> dict:
    """Value of a given field on the target kind to convert is either an input source attribute/relationship of the source node,
    or a raw value."""

    data = {}
    for dest_field_name, input_for_dest_field in mapping.items():
        value = input_for_dest_field.value
        if isinstance(value, str):  # source_field
            item = getattr(node, value)
            if isinstance(item, BaseAttribute):
                data[dest_field_name] = item.value
            elif isinstance(item, RelationshipManager):
                if item.schema.cardinality == RelationshipCardinality.ONE:
                    peer = await item.get_peer(db=db)
                    if peer is not None:
                        data[dest_field_name] = {"id": peer.id}
                    # else, relationship is optional, and if the target relationship is mandatory an error will be raised during creation
                elif item.schema.cardinality == RelationshipCardinality.MANY:
                    data[dest_field_name] = [{"id": peer.id} for _, peer in (await item.get_peers(db=db)).items()]
                else:
                    raise ValueError(f"Unknown cardinality {item.schema.cardinality=}")
        else:  # user input data
            data[dest_field_name] = value.value
    return data


async def get_unidirectional_rels_peers_ids(node: Node, branch: Branch, db: InfrahubDatabase) -> list[str]:
    """
    Returns peers ids of nodes connected to input `node` through an incoming unidirectional relationship.
    """

    out_rels_identifier = [rel.identifier for rel in node.get_schema().relationships]
    query = await GetAllPeersIds.init(db=db, node_id=node.id, branch=branch, exclude_identifiers=out_rels_identifier)
    await query.execute(db=db)
    return query.get_peers_uuids()


async def convert_object_type(
    node: Node, target_schema: NodeSchema, mapping: dict[str, InputForDestField], branch: Branch, db: InfrahubDatabase
) -> Node:
    """Delete the node and return the new created one. If creation fails, the node is not deleted, and raise an error.
    An extra check is performed on input node peers relationships to make sure they are still valid."""

    node_schema = node.get_schema()
    if not isinstance(node_schema, NodeSchema):
        raise ValueError(f"Only a node with a NodeSchema can be converted, got {type(node_schema)}")

    async with db.start_transaction() as dbt:
        deleted_node_out_rels_peer_ids = await get_out_rels_peers_ids(node=node, db=dbt)
        deleted_node_unidir_rels_peer_ids = await get_unidirectional_rels_peers_ids(node=node, db=dbt, branch=branch)

        # Delete the node, so we delete relationships with peers as well, which might temporarily break cardinality constraints
        # but they should be restored when creating the new node.
        deleted_nodes = await NodeManager.delete(db=dbt, branch=branch, nodes=[node], cascade_delete=False)
        if len(deleted_nodes) != 1:
            raise ValueError(f"Deleted {len(deleted_nodes)} nodes instead of 1")

        data_new_node = await build_data_new_node(dbt, mapping, node)
        node_created = await create_node(
            data=data_new_node,
            db=dbt,
            branch=branch,
            schema=target_schema,
        )

        # Make sure relationships with constraints are not broken by retrieving them
        peers_ids = deleted_node_out_rels_peer_ids + deleted_node_unidir_rels_peer_ids
        peers = await NodeManager.get_many(ids=peers_ids, db=dbt, prefetch_relationships=True, branch=branch)
        for peer in peers.values():
            peer.validate_relationships()

        # If the node had some value reserved in any Pools / Resource Manager, we need to change the identifier of the reservation(s)
        query = await PoolChangeReserved.init(
            db=dbt,
            existing_identifier=node.get_id(),
            new_identifier=node_created.get_id(),
            branch=branch,
        )
        await query.execute(db=dbt)

        return node_created



from infrahub.permissions.backend import PermissionBackend
from infrahub.permissions.globals import define_global_permission_from_branch
from infrahub.permissions.local_backend import LocalPermissionBackend
from infrahub.permissions.manager import PermissionManager
from infrahub.permissions.report import report_schema_permissions
from infrahub.permissions.types import AssignedPermissions, get_global_permission_for_kind

__all__ = [
    "AssignedPermissions",
    "LocalPermissionBackend",
    "PermissionBackend",
    "PermissionManager",
    "define_global_permission_from_branch",
    "get_global_permission_for_kind",
    "report_schema_permissions",
]



from __future__ import annotations

import importlib
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from infrahub.config import CacheDriver
    from infrahub.message_bus.types import KVTTL


class InfrahubCache(ABC):
    """Base class for caching services"""

    @abstractmethod
    async def delete(self, key: str) -> None:
        """Delete a key from the cache."""
        raise NotImplementedError()

    @abstractmethod
    async def get(self, key: str) -> str | None:
        """Retrieve a value from the cache."""
        raise NotImplementedError()

    @abstractmethod
    async def get_values(self, keys: list[str]) -> list[str | None]:
        """Return a list the values for requested keys."""
        raise NotImplementedError()

    @abstractmethod
    async def list_keys(self, filter_pattern: str) -> list[str]:
        """Return a list of active keys that match the provided filter."""
        raise NotImplementedError()

    @abstractmethod
    async def set(self, key: str, value: str, expires: KVTTL | None = None, not_exists: bool = False) -> bool | None:
        """Set a value in the cache."""
        raise NotImplementedError()

    @classmethod
    async def new_from_driver(cls, driver: CacheDriver) -> InfrahubCache:
        """Imports and initializes the correct class based on the supplied driver.

        This is to ensure that we only import the Python modules that we actually
        need to operate and not import all possible options.
        """
        module = importlib.import_module(driver.driver_module_path)
        broker_driver: InfrahubCache = getattr(module, driver.driver_class_name)
        return await broker_driver.new()

    @classmethod
    async def new(cls) -> InfrahubCache:
        raise NotImplementedError()



from __future__ import annotations

from pydantic import BaseModel, Field


class ContextAccount(BaseModel):
    id: str = Field(..., description="The ID of the account")


class RequestContext(BaseModel):
    """The context can be used to override settings such as the account within mutations."""

    account: ContextAccount | None = Field(default=None, description="Account tied to the context")



import re
import sys

import typer
from infrahub_sdk import Config, InfrahubClientSync

from infrahub import config
from infrahub.core.constants import InfrahubKind

app = typer.Typer()

REGEX_USERNAME = r"^Username.*\'(.*)\'"
REGEX_PASSWORD = r"^Password.*\'(.*\:\/\/)(.*)@(.*)\'"


@app.command()
def askpass(
    text: list[str] | None = typer.Argument(None),
    config_file: str = typer.Option("infrahub.toml", envvar="INFRAHUB_CONFIG"),
) -> None:
    config.SETTINGS.initialize_and_exit(config_file=config_file)

    text = text or sys.stdin.read().strip()
    request_type = None

    if isinstance(text, list):
        text = " ".join(text)

    if re.match(pattern=REGEX_USERNAME, string=text):
        location = re.search(REGEX_USERNAME, text).group(1)
        request_type = "username"
    elif re.match(pattern=REGEX_PASSWORD, string=text):
        location = f"{re.search(REGEX_PASSWORD, text).group(1)}{re.search(REGEX_PASSWORD, text).group(3)}"
        request_type = "password"

    if not request_type:
        raise typer.Exit(f"Unable to identify the request type in '{text}'")

    client = InfrahubClientSync(config=Config(address=config.SETTINGS.main.internal_address, insert_tracker=True))
    repo = client.get(kind=InfrahubKind.GENERICREPOSITORY, location__value=location)

    if not repo.credential._id:
        raise typer.Exit("Repository doesn't have credentials defined.")

    repo.credential.fetch()

    attr = getattr(repo.credential.peer, request_type)
    print(attr.value)



import os
import subprocess  # noqa: S404
import uuid
import warnings
from pathlib import Path

import pytest
from prefect.client.orchestration import PrefectClient

from infrahub_testcontainers import __version__ as infrahub_version

from .container import PROJECT_ENV_VARIABLES, InfrahubDockerCompose


class TestInfrahubDocker:
    @pytest.fixture(scope="class")
    def infrahub_version(self) -> str:
        return os.getenv("INFRAHUB_TESTING_IMAGE_VER") or infrahub_version

    @staticmethod
    def execute_ctl_run(address: str, script: str) -> str:
        env = os.environ.copy()
        env["INFRAHUB_ADDRESS"] = address
        env["INFRAHUB_API_TOKEN"] = PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_INITIAL_ADMIN_TOKEN"]
        env["INFRAHUB_MAX_CONCURRENT_EXECUTION"] = "1"
        result = subprocess.run(  # noqa: S602
            f"infrahubctl run {script}", shell=True, capture_output=True, text=True, env=env, check=False
        )
        return result.stdout

    @staticmethod
    def execute_command(command: str, address: str, concurrent_execution: int = 10) -> subprocess.CompletedProcess[str]:
        env = os.environ.copy()
        env["INFRAHUB_ADDRESS"] = address
        env["INFRAHUB_API_TOKEN"] = PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_INITIAL_ADMIN_TOKEN"]
        env["INFRAHUB_MAX_CONCURRENT_EXECUTION"] = f"{concurrent_execution}"
        result = subprocess.run(  # noqa: S602
            command, shell=True, capture_output=True, text=True, env=env, check=False
        )
        return result

    @pytest.fixture(scope="class")
    def tmp_directory(self, tmpdir_factory: pytest.TempdirFactory) -> Path:
        name = f"{self.__class__.__name__.lower()}_{uuid.uuid4().hex}"
        return Path(str(tmpdir_factory.mktemp(name)))

    @pytest.fixture(scope="class")
    def remote_repos_dir(self, tmp_directory: Path) -> Path:
        directory = tmp_directory / PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_LOCAL_REMOTE_GIT_DIRECTORY"]
        directory.mkdir(exist_ok=True)

        return directory

    @pytest.fixture(scope="class")
    def remote_backups_dir(self, tmp_directory: Path) -> Path:
        directory = tmp_directory / PROJECT_ENV_VARIABLES["INFRAHUB_TESTING_LOCAL_DB_BACKUP_DIRECTORY"]
        directory.mkdir(exist_ok=True)

        return directory

    @pytest.fixture(scope="class")
    def default_branch(self) -> str:
        return "main"

    @pytest.fixture(scope="class")
    def deployment_type(self, request: pytest.FixtureRequest) -> str | None:
        return request.config.getoption(name="infrahub_deployment_type", default=None)

    @pytest.fixture(scope="class")
    def infrahub_compose(
        self,
        tmp_directory: Path,
        remote_repos_dir: Path,  # initialize repository before running docker compose to fix permissions issues # noqa: ARG002
        remote_backups_dir: Path,  # noqa: ARG002
        infrahub_version: str,
        deployment_type: str | None,
    ) -> InfrahubDockerCompose:
        return InfrahubDockerCompose.init(
            directory=tmp_directory, version=infrahub_version, deployment_type=deployment_type
        )

    @pytest.fixture(scope="class")
    def infrahub_app(self, request: pytest.FixtureRequest, infrahub_compose: InfrahubDockerCompose) -> dict[str, int]:
        tests_failed_before_class = request.session.testsfailed

        def cleanup() -> None:
            tests_failed_during_class = request.session.testsfailed - tests_failed_before_class
            if tests_failed_during_class > 0:
                stdout, stderr = infrahub_compose.get_logs("infrahub-server", "task-worker")
                warnings.warn(f"Container logs:\nStdout:\n{stdout}\nStderr:\n{stderr}", stacklevel=2)
            infrahub_compose.stop()

        request.addfinalizer(cleanup)

        try:
            infrahub_compose.start()
        except Exception as exc:
            stdout, stderr = infrahub_compose.get_logs()
            raise Exception(f"Failed to start docker compose:\nStdout:\n{stdout}\nStderr:\n{stderr}") from exc

        return infrahub_compose.get_services_port()

    @pytest.fixture(scope="class")
    def infrahub_port(self, infrahub_app: dict[str, int]) -> int:
        return infrahub_app["server"]

    @pytest.fixture(scope="class")
    def task_manager_port(self, infrahub_app: dict[str, int]) -> int:
        return infrahub_app["task-manager"]

    @pytest.fixture(scope="class")
    def prefect_client(self, task_manager_port: int) -> PrefectClient:
        return PrefectClient(api=f"http://localhost:{task_manager_port}/api/")



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field

from .node import InfrahubNode  # noqa: TC001


class RepositoryBranchInfo(BaseModel):
    internal_status: str


class RepositoryData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    repository: InfrahubNode = Field(..., description="InfrahubNode representing a Repository")
    branches: dict[str, str] = Field(
        ...,
        description="Dictionary with the name of the branch as the key and the active commit id as the value",
    )

    branch_info: dict[str, RepositoryBranchInfo] = Field(default_factory=dict)

    def get_staging_branch(self) -> str | None:
        for branch, info in self.branch_info.items():
            if info.internal_status == "staging":
                return branch
        return None



from typing import Callable

from infrahub.auth import AccountSession
from infrahub.core.branch import Branch
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import AuthorizationError
from infrahub.graphql.analyzer import InfrahubGraphQLQueryAnalyzer
from infrahub.graphql.initialization import GraphqlParams

from .interface import CheckerResolution, GraphQLQueryPermissionCheckerInterface


class AnonymousGraphQLPermissionChecker(GraphQLQueryPermissionCheckerInterface):
    def __init__(self, anonymous_access_allowed_func: Callable[[], bool]) -> None:
        self.anonymous_access_allowed_func = anonymous_access_allowed_func

    async def supports(self, db: InfrahubDatabase, account_session: AccountSession, branch: Branch) -> bool:  # noqa: ARG002
        return not account_session.authenticated

    async def check(
        self,
        db: InfrahubDatabase,  # noqa: ARG002
        account_session: AccountSession,  # noqa: ARG002
        analyzed_query: InfrahubGraphQLQueryAnalyzer,
        query_parameters: GraphqlParams,  # noqa: ARG002
        branch: Branch,  # noqa: ARG002
    ) -> CheckerResolution:
        if not self.anonymous_access_allowed_func() or analyzed_query.contains_mutation:
            raise AuthorizationError("Authentication is required to perform this operation")
        return CheckerResolution.NEXT_CHECKER



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from graphene import Boolean, InputObjectType, Mutation, String

from infrahub.core.account import ObjectPermission
from infrahub.core.constants import ComputedAttributeKind, PermissionAction, PermissionDecision
from infrahub.core.manager import NodeManager
from infrahub.core.registry import registry
from infrahub.database import retry_db_transaction
from infrahub.events import EventMeta
from infrahub.events.node_action import NodeUpdatedEvent
from infrahub.exceptions import NodeNotFoundError, ValidationError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.types.context import ContextInput
from infrahub.log import get_log_data
from infrahub.worker import WORKER_IDENTITY

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.graphql.initialization import GraphqlContext


class InfrahubComputedAttributeUpdateInput(InputObjectType):
    id = String(required=True)
    kind = String(required=True)
    attribute = String(required=True)
    value = String(required=True)


class UpdateComputedAttribute(Mutation):
    class Arguments:
        data = InfrahubComputedAttributeUpdateInput(required=True)
        context = ContextInput(required=False)

    ok = Boolean()

    @classmethod
    @retry_db_transaction(name="update_computed_attribute")
    async def mutate(
        cls,
        _: dict,
        info: GraphQLResolveInfo,
        data: InfrahubComputedAttributeUpdateInput,
        context: ContextInput | None = None,
    ) -> UpdateComputedAttribute:
        graphql_context: GraphqlContext = info.context
        node_schema = registry.schema.get_node_schema(
            name=str(data.kind), branch=graphql_context.branch.name, duplicate=False
        )
        target_attribute = node_schema.get_attribute(name=str(data.attribute))
        if (
            not target_attribute.computed_attribute
            or target_attribute.computed_attribute.kind == ComputedAttributeKind.USER
        ):
            raise ValidationError(input_value=f"{node_schema.kind}.{target_attribute.name} is not a computed attribute")

        graphql_context.active_permissions.raise_for_permission(
            permission=ObjectPermission(
                namespace=node_schema.namespace,
                name=node_schema.name,
                action=PermissionAction.UPDATE.value,
                decision=PermissionDecision.ALLOW_DEFAULT.value
                if graphql_context.branch.name == registry.default_branch
                else PermissionDecision.ALLOW_OTHER.value,
            )
        )
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        if not (
            target_node := await NodeManager.get_one(
                db=graphql_context.db,
                kind=node_schema.kind,
                id=str(data.id),
                branch=graphql_context.branch,
                fields={target_attribute.name: None},
            )
        ):
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated node was not found in the database",
            )

        attribute_field = getattr(target_node, str(data.attribute), None)
        if not attribute_field:
            raise NodeNotFoundError(
                node_type="target_node",
                identifier=str(data.id),
                message="The indicated not does not have the specified attribute_name",
            )
        if attribute_field.value != str(data.value):
            attribute_field.value = str(data.value)
            async with graphql_context.db.start_transaction() as dbt:
                await target_node.save(db=dbt, fields=[str(data.attribute)])

            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            event = NodeUpdatedEvent(
                kind=node_schema.kind,
                node_id=target_node.get_id(),
                changelog=target_node.node_changelog.model_dump(),
                fields=[str(data.attribute)],
                meta=EventMeta(
                    context=graphql_context.get_context(),
                    initiator_id=WORKER_IDENTITY,
                    request_id=request_id,
                    account_id=graphql_context.active_account_session.account_id,
                    branch=graphql_context.branch,
                ),
            )
            await graphql_context.active_service.event.send(event=event)

        result: dict[str, Any] = {"ok": True}

        return cls(**result)



from prefect import flow
from prefect.client.orchestration import get_client

from infrahub.actions.gather import gather_trigger_action_rules
from infrahub.computed_attribute.gather import (
    gather_trigger_computed_attribute_jinja2,
    gather_trigger_computed_attribute_python,
)
from infrahub.services import InfrahubServices
from infrahub.trigger.catalogue import builtin_triggers
from infrahub.webhook.gather import gather_trigger_webhook

from .setup import setup_triggers


@flow(name="trigger-configure-all", flow_run_name="Configure all triggers")
async def trigger_configure_all(service: InfrahubServices) -> None:
    async with service.database.start_session() as db:
        webhook_trigger = await gather_trigger_webhook(db=db)
        computed_attribute_j2_triggers = await gather_trigger_computed_attribute_jinja2()
        (
            computed_attribute_python_triggers,
            computed_attribute_python_query_triggers,
        ) = await gather_trigger_computed_attribute_python(db=db)
        action_rules = await gather_trigger_action_rules(db=db)
        triggers = (
            computed_attribute_j2_triggers
            + computed_attribute_python_triggers
            + computed_attribute_python_query_triggers
            + builtin_triggers
            + webhook_trigger
            + action_rules
        )

        async with get_client(sync_client=False) as prefect_client:
            await setup_triggers(client=prefect_client, triggers=triggers, force_update=True)



from infrahub.core.validators.attribute.length import AttributeLengthChecker
from infrahub.dependencies.interface import DependencyBuilder, DependencyBuilderContext


class SchemaAttributLengthConstraintDependency(DependencyBuilder[AttributeLengthChecker]):
    @classmethod
    def build(cls, context: DependencyBuilderContext) -> AttributeLengthChecker:
        return AttributeLengthChecker(db=context.db, branch=context.branch)



from __future__ import annotations

from infrahub_sdk.graphql import Mutation
from prefect import flow

from infrahub.context import InfrahubContext  # noqa: TC001  needed for prefect flow
from infrahub.services import InfrahubServices  # noqa: TC001  needed for prefect flow
from infrahub.trigger.models import TriggerType
from infrahub.trigger.setup import setup_triggers_specific
from infrahub.workflows.utils import add_tags

from .gather import gather_trigger_action_rules
from .models import EventGroupMember  # noqa: TC001  needed for prefect flow


@flow(
    name="action-add-node-to-group",
    flow_run_name="Adding node={node_id} to group={group_id}",
)
async def add_node_to_group(
    branch_name: str,
    node_id: str,
    group_id: str,
    context: InfrahubContext,  # noqa: ARG001
    service: InfrahubServices,
) -> None:
    await add_tags(branches=[branch_name], nodes=[node_id, group_id])

    mutation = Mutation(
        mutation="RelationshipAdd",
        input_data={"data": {"id": group_id, "name": "members", "nodes": [{"id": node_id}]}},
        query={"ok": None},
    )

    await service.client.execute_graphql(query=mutation.render(), branch_name=branch_name)


@flow(
    name="action-remove-node-from-group",
    flow_run_name="Removing node={node_id} from group={group_id}",
)
async def remove_node_from_group(
    branch_name: str,
    node_id: str,
    group_id: str,
    context: InfrahubContext,  # noqa: ARG001
    service: InfrahubServices,
) -> None:
    await add_tags(branches=[branch_name], nodes=[node_id, group_id])

    mutation = Mutation(
        mutation="RelationshipRemove",
        input_data={"data": {"id": group_id, "name": "members", "nodes": [{"id": node_id}]}},
        query={"ok": None},
    )

    await service.client.execute_graphql(query=mutation.render(), branch_name=branch_name)


@flow(
    name="action-run-generator",
    flow_run_name="Running generator generator_definition_id={generator_definition_id} for nodes={node_ids}",
)
async def run_generator(
    branch_name: str,
    node_ids: list[str],
    generator_definition_id: str,
    context: InfrahubContext,  # noqa: ARG001
    service: InfrahubServices,
) -> None:
    await add_tags(branches=[branch_name], nodes=node_ids + [generator_definition_id])
    await _run_generator(
        branch_name=branch_name, generator_definition_id=generator_definition_id, node_ids=node_ids, service=service
    )


@flow(
    name="action-run-generator-group-event",
    flow_run_name="Running generator",
)
async def run_generator_group_event(
    branch_name: str,
    members: list[EventGroupMember],
    generator_definition_id: str,
    context: InfrahubContext,  # noqa: ARG001
    service: InfrahubServices,
) -> None:
    node_ids = [node.id for node in members]
    await add_tags(branches=[branch_name], nodes=node_ids + [generator_definition_id])
    await _run_generator(
        branch_name=branch_name, generator_definition_id=generator_definition_id, node_ids=node_ids, service=service
    )


@flow(
    name="configure-action-rules",
    flow_run_name="Configure updated action rules and triggers",
)
async def configure_action_rules(
    service: InfrahubServices,
) -> None:
    await setup_triggers_specific(
        gatherer=gather_trigger_action_rules, trigger_type=TriggerType.ACTION_TRIGGER_RULE, db=service.database
    )  # type: ignore[misc]


async def _run_generator(
    branch_name: str,
    node_ids: list[str],
    generator_definition_id: str,
    service: InfrahubServices,
) -> None:
    mutation = Mutation(
        mutation="CoreGeneratorDefinitionRun",
        input_data={"data": {"id": generator_definition_id, "nodes": node_ids}},
        query={"ok": None},
    )

    await service.client.execute_graphql(query=mutation.render(), branch_name=branch_name)
